#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
Regression trees are decision trees adapted to the regression task, which store
numeric target attribute values instead of class labels in leaves, and use
appropriately modified split selection and stop criteria.

As with decision trees, regression tree nodes decompose the data into subsets,
and regression tree leaves correspond to sufficiently small or sufficiently
uniform subsets. Splits are selected to decrease the dispersion of target
attribute values, so that they can be reasonably well predicted by their mean
values at leaves. The resulting model is piecewise-constant, with fixed
predicted values assigned to regions to which the domain is decomposed by
the tree structure.
"""
from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.predictive.regression import Regression
from nzpyida.analytics.utils import map_to_props

class DecisionTreeRegressor(Regression):
    """
    Decision tree based regressor
    """
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the regressor class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.fit_proc = 'REGTREE'
        self.predict_proc = 'PREDICT_REGTREE'
        self.target_column_in_output = "CLASS"
        self.has_print_proc = True

    def fit(self, in_df: IdaDataFrame, target_column: str, id_column: str=None,
        in_columns: List[str]=None, col_def_type: str=None, col_def_role: str=None,
        col_properties_table: str=None, eval_measure: str=None, min_improve: float=0.1,
        min_split: int=50, max_depth: int=10, val_table: str=None, qmeasure: str=None,
        statistics: str=None):
        """
        This function creates a regression tree model based on provided data and
        store it in a database.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame

        target_column : str
            the input table column representing the prediction target, definition of multitargets
            can be processed by 'incolumn' parameter and column properties.

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        nominal_colums : str, optional
            the input table nominal columns, if any, separated by a semi-colon (;).
            Parameter 'nominalCols' is deprecated please use 'incolumn' intead.

        in_columns : List[str], optional
            the list of input table columns with special properties.
            Each column is followed by one or several of the following properties:
                its type: ':nom' (for nominal), ':cont' (for continuous).
                Per default, all numerical types are con-tinuous, other types are nominal.
                its role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same
            as ':colweight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns. Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input table column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight' is
            unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsup-ported,
            i.e. same as '1')

        eval_measure : str, optional
            the split evaluation measure. Allowed values are: variance.

        min_improve : float, optional
            the minimum improvement of the split evaluation measure required

        min_split : int, optional
            the minimum number of instances per tree node that can be split

        max_depth : int, optional
            the maximum number of tree levels (including leaves)

        val_table : str, optional
            the input table containing the validation dataset.
            If this parameter is undefined, no pruning will be performed.

        qmeasure : str, optional
            the quality measure for pruning the tree. Allowed values are: mse, r2.

        statistics : str, optional
            flags indicating which statistics to collect.
            Allowed values are: none, columns, values:n, all.
            If statistics=none, no statistics are collected.
            If statistics=columns, statistics on the input table columns like mean value
            are collected.
            If statistics=values:n with n a positive number, statistics about the columns
            and the column val-ues are collected. Up to <n> column value statistics
            are collected:
            If a nominal column contains more than <n> values, only the <n> most frequent
            column stat-istics are kept.
            If a numeric column contains more than <n> values, the values will be discretized
            and the stat-istics will be collected on the discretized values.
            Indicating statistics=all is equal to statistics=values:100.

        """

        params = {
            'id': id_column,
            'target': target_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
            'eval': eval_measure,
            'minimprove': min_improve,
            'minsplit': min_split,
            'maxdepth': max_depth,
            'valtable': val_table,
            'qmeasure': qmeasure,
            'statistics': statistics
        }

        self._fit(in_df=in_df, params=params)

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None, variance: bool=False):
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame to predict

        out_table : str, optional
            the output table where the predictions will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model

        variance : bool, optional
            a flag indicating whether the variance of the predictions should be included
            into the output table

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """

        params = {
            'id': id_column,
            'var': variance
            }

        return self._predict(in_df=in_df, params=params, out_table=out_table)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
Association rules mining is a popular method for discovering interesting 
and useful patterns in a large scale transaction database. The database 
contains transactions which consist of a set of items and a transaction 
identifier (e.g., a market basket). Association rules are implications 
of the form X -> Y where X and Y are two disjoint subsets of all available 
items. X is called the antecedent or LHS (left hand side) and Y is called 
the consequent or RHS (right hand side). Discovered association rules have 
to satisfy user-defined constraints on measures of significance and interest.

The Apriori algorithm organizes the search for frequent itemsets by systematically 
considering itemsets of increasing size in consecutive iterations. Due to its method 
of calculation, the number of candidates identified by the Apriori algorithm may 
be overwhelming for extremely large data sets or a low support threshold. Because 
of this limitation, the FP-growth algorithm is provided in the IBM Netezza In-Database 
Analytics package instead.

The FP-growth algorithm avoids candidate generation as well as multiple passes through 
the data by creating a data structure called a frequent pattern tree, or FP-tree. 
This tree is a compact representation of the data set contents sufficient for finding 
frequent itemsets. Nodes of the tree represent single items and store their occurrence 
counts. Only items with sufficiently high support, frequent item-sets of size 1, are 
represented. Branches, called node-links, in the FP-tree connect nodes that represent 
items co-occurring for some instances in the data set. There is also a frequent item 
header table that points to nodes corresponding to particular items.

The tree is built by identifying all frequent items and their counts, then consecutively 
“inserting” each transaction to the tree. This requires exactly two scans of the 
data set, regardless of its size or support threshold level. The FP-tree is used to 
identity frequent itemsets using a frequent pattern growth process, which traverses 
the tree by following node-links in an appropriate way.

By avoiding explicit candidate generation, the FP-growth algorithm reduces the number 
of data set scans. It can also perform efficiently, regardless of the threshold support.
"""

from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling

class ARule(PredictiveModeling):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates Association Rules Class

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - the name of the Association Rules model to build
        """
        super().__init__(idadb, model_name)
        self.has_print_proc = True
        self.fit_proc = 'ARULE'
        self.predict_proc = 'PREDICT_ARULE'
        self.model_name = model_name
    
    def fit(self, in_df: IdaDataFrame, transaction_id_column: str='tid', item_column: str='item', by_column: str=None, 
            level: int=1, max_set_size: int=6, support: float=None, support_type: str='percent',
            confidence: float=0.5):
        """
        This function builds an Association Rules model. The model is saved to the 
        database in a set of tables and registered in the database model metadata. Use the 
        function 'describe' to display the Association Rules of the model, or the 
        Model Management functions to further manipulate the model

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        transaction_id_column : str, optional
            the input table column identifying transactions
        
        items_column : str, optional
            the input table column identifying items in transactions

        by_column : str, optional
            the input table column identifying groups of transactions if any. 
            Association Rules min-ing is done separately on each of these groups. 
            Leave the parameter undefined if no groups are to be considered.
        
        level : int, optional
            ARULE first temporarily redistributes the data into overlapping parts in such a way, 
            that each part can be processed in parallel without communication between the SPUs. 
            Note that for this to work, there can be redundancy between the parts, such that 
            the accumulative size of the temporary parts can be much higher than the one of 
            the ori- ginal data set. The parameter lvl controls how many parts are created. 
            The higher lvl: 
                - The more computation and temporary database space is required for the splitting
                - The smaller the amount of main memory that is required for each data slice

            Note: To fully use the benefits of parallel computing, do not specify the value 
            of the lvl parameter too low. Additionally, the lower the value of the lvl parameter,
            the higher the memory consumption for each part. The higher memory consumption might 
            cause an out-of-memory error on the SPUs. If an out-of-memory error on the SPUs oc-curs, 
            increase the lvl parameter.
            
            If you specify the value 0, the algorithm is executed in a serial way for each data 
            set group. However, only if the data set fits in one node, and only if the splitting 
            increases the total number of rows dramatically, the stored procedure might be executed 
            faster when you specify the value 0.
            Default - 1
            Min - 0
        
        support, int, optional
            minimum support value satisfied by all association rules. According to supporttype, 
            it defines the absolute number (#supporting transactions) or the percentage of 
            transactions (#supporting transactions/#total transactions*100). Too low minimum support 
            increases the number of generated rules and the computational expense.
        
        support_type: str, optional
            the type how the minimum support should be interpreted. The following values are allowed: 
            absolute, percent.
            Note the support and support_type values are common to all groups in the dataset. 
            E.g. if 3 is the absolute minimum support, then an itemset will be considered frequent 
            if at least 3 trans-actions contain its items, no matter what is the number of transactions 
            in this group. Use support_type='percent' to indicate a minimum support depending on the 
            size of the groups. Specifying support_type='absolute' takes effect only if a support is 
            explicitly supplied.

        confidence: float, optional
            the minimum confidence for an association rule to be
            default - 0.5
            min - 0
            max - 1 
        """
        if support_type == 'percent' and not support:
            support = 5.0
        params = {
            'tid': transaction_id_column,
            'item': item_column,
            'by': by_column,
            'lvl': level,
            'maxsetsize': max_set_size,
            'support': support,
            'supporttype': support_type,
            'confidence': confidence
        }

        self._fit(in_df=in_df, params=params, needs_id=False)
    
    def predict(self, in_df: IdaDataFrame, out_table: str=None, transaction_id_column: str='tid', item_column: str='item',
                by_column: str=None, scoring_type: str='exclusiveRecommend', name_map_column: str=None, item_name_column: str='item',
                item_name_mapped_column: str='item_name', min_size: int=1, max_size: int=64, min_support: float=0.0, max_support: float=1.0,
                min_confidence: float=0.0, max_confidence: float=1.0, min_lift: float=None, max_lift: float=None, 
                min_conviction: float=0.0, max_conviction: float=None, min_affinity: float=0.0, max_affinity: float=1.0,
                min_leverage: float=-0.25, max_leverage: float=1.0) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table where the predictions will be stored
        
        transaction_id_column : str, optional
            the input table column identifying transactions
        
        items_column : str, optional
            the input table column identifying items in transactions

        by_column : str, optional
            the input table column identifying groups of transactions if any. 
            Association Rules min-ing is done separately on each of these groups. 
            Leave the parameter undefined if no groups are to be considered.

        scoring_type: str, optional
            he type how the scoring algorithm should be applied to the input data. 
            The following values are allowed: recommend, exclusiveRecommend.
            recommend - A rule is returned if its left hand side itemset is a subset of the transaction.
            exclusiveRecommend - A rule is returned if its left hand side itemset is a subset of 
            the input itemset, and its right hand side itemset is not a subset of the transaction.
        
        name_map_column: str, optional
            table which provides names of items and their associated mapped values 
            in LHS_ITEMS, RHS_ITEMS columns of outtable

        item_name_column, str, optional
            the column name of namemap table where the item identifiers are

        item_name_mapped_column: str, optional
            the column name of namemap table where the item names are stored which should be used in-stead of the item identifier

        min_size: int, optional
            The minimum number of items per association rule to be applied

        max_size: int, optional
            The maximum number of items per association rule to be applied

        min_support: float, optional
            The minimum support of an association rule to be applied

        max_support: float, optional
            The maximum support of an association rule to be applied

        min_confidence: float, optional
            The minimum confidence of an association rule to be applied.

        max_confidence: float, optional
            The maximum confidence of an association rule to be applied

        min_lift: float, optional
            The minimum lift of an association rule to be applied

        max_lift: float, optional
            The maximum lift of an association rule to be applied

        min_conviction: float, optional
            The minimum conviction of an association rule to be applied

        max_conviction: float, optional
            The maximum conviction of an association rule to be applied

        min_affinity: float, optional
            The minimum affinity of an association rule to be applied
        
        max_affinity: float, optional
            The maximum affinity of an association rule to be applied

        min_leverage: float, optional
            The minimum leverage of an association rule to be applied

        max_leverage: float, optional
            The maximum leverage of an association rule to be applied

        Returns
        -------
        IdaDataFrame
            the data frame containing output of a Association Rules model prediction
        """
        params = {
            'tid': transaction_id_column,
            'item': item_column,
            'by': by_column,
            'type': scoring_type,
            'namemap': name_map_column,
            'itemname': item_name_column,
            'itemnamemapped': item_name_mapped_column,
            'minsize': min_size,
            'maxsize': max_size,
            'minsupp': min_support,
            'maxsupp': max_support,
            'minconf': min_confidence,
            'maxconf': max_confidence,
            'minlift': min_lift,
            'maxlift': max_lift,
            'minconv': min_conviction,
            'maxconv': max_conviction,
            'minaffi': min_affinity,
            'maxaffi': max_affinity,
            'minleve': min_leverage,
            'maxleve': max_leverage
        }
        return self._predict(in_df=in_df, params=params, out_table=out_table)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
In many classification applications it may be required or desirable not
only to accurately classify instances, but also to inspect the model.
The inspection makes it possible to explain its decisions, modify it,
or combine with some existing background knowledge. In such applications,
where both the high classification accuracy and human-readability of the
model are required, the method of choice is typically going to be
decision trees.

A decision tree is a hierarchical structure that represents a classification
model using a "divide and conquer" approach. Internal tree nodes represent
splits applied to decompose the data set into subsets, and terminal nodes,
also referred to as leaves, assign class labels to sufficiently small
or uniform subsets. Splits are specified by logical conditions based on
selected single attributes, with a separate outgoing branch corresponding
to each possible outcome.

The concept of decision tree construction is to select splits that decrease
the impurity of class distribution in the resulting subsets of instances,
and increase the domination of one or more classes over the others.
The goal is to find a subset containing only or mostly instances of one
class after a small number of splits, so that a leaf with that class
label is created. This approach promotes simple trees, which typically
generalize better.
"""

from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.predictive.classification import Classification


class DecisionTreeClassifier(Classification):
    """
    Decision tree based classifier.
    """

    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the classifier class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.fit_proc = 'DECTREE'
        self.predict_proc = 'PREDICT_DECTREE'
        self.has_print_proc = True

    def fit(self, in_df: IdaDataFrame, target_column: str, id_column: str=None,
        in_columns: List[str]=None, col_def_type: str=None, col_def_role: str=None,
        col_properties_table: str=None, weights: str=None, eval_measure: str=None,
        min_improve: float=0.02, min_split: int=50, max_depth: int=10, val_table: str=None,
        val_weights: str=None, qmeasure: str=None, statistics: str=None):
        """
        Grows the decision tree and stores its model in the database.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        in_columns : str, optional
            the list of input table columns with special properties.
            Each column is followed by one or several of the following properties:
                its type: ':nom' (for nominal), ':cont' (for continuous).
                    Per default, all numerical types are continuous, other types are nominal.
                its role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same as
            ':col-weight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns. Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            The format of this table is the output format of stored procedure 
            nza..COLUMN_PROPERTIES(). If the parameter is undefined, the input table column
            properties will be detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight' is unsupported,
            i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')

        weights : str, optional
            the input table containing optional instance or class weights for the input 
            table columns.
            If the parameter is undefined, we assume that the weights are uniformly equal to 1.
            The <weights> table contains following columns:
                weight: a numeric column containing the instance or class weight,
                id: a column to be joined with the <id> column of <intable>, defining 
                instance weights,
                class: a column to be joined with the <target> column of <intable>, defining 
                class weights.
            The id or class column can be missing, at least one of them must be present.
            For instances or classes not occurring in this table, weights of 1 are assumed.

        eval_measure : str, optional
            the class impurity measure used for split evaluation. 
            Allowed values are 'entropy' and 'gini'

        min_improve : float, optional
            the minimum improvement of the split evaluation measure required

        min_split : int, optional
            the minimum number of instances per tree node that can be split

        max_depth : int, optional
            the maximum number of tree levels (including leaves)

        val_table : str, optional
            the input table containing the validation dataset.
            If this parameter is undefined, no pruning will be performed.

        val_weights : str, optional
            the input table containing optional instance or class weights for the 
            validation dataset.
            It is similar to the <weights> table.

        qmeasure : str, optional
            the quality measure for pruning.
            Allowed values are Acc or wAcc.

        statistics : str, optional
            flags indicating which statistics to collect.
            Allowed values are: none, columns, values:n, all.
            If statistics=none, no statistics are collected.
            If statistics=columns, statistics on the input table columns like mean 
            value are collected.
            If statistics=values:n with n a positive number, statistics about the 
            columns and the column
            values are collected. Up to <n> column value statistics are collected:
                If a nominal column contains more than <n> values, only the <n> most 
                frequent column statistics are kept.
                If a numeric column contains more than <n> values, the values will be
                discretized and the statistics will be collected on the discretized values.
            Indicating statistics=all is equal to statistics=values:100.
        """

        params = {
            'id': id_column,
            'target': target_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
            'weights': weights,
            'eval': eval_measure,
            'minimprove': min_improve,
            'minsplit': min_split,
            'maxdepth': max_depth,
            'valtable': val_table,
            'valweights': val_weights,
            'qmeasure': qmeasure,
            'statistics': statistics
        }

        self._fit(in_df=in_df, params=params)

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None, prob: bool=False, out_table_prob: str=None) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table where the predictions will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model

        prob : bool, optional
            the flag indicating whether the probability of the predicted class should be included
            into the output table or not

        out_table_prob : str, optional
            if specified, the probability output table where class probability predictions
            will be stored

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """

        params = {
            'id': id_column,
            'prob': prob,
            'outtableprob': out_table_prob
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import threading
from nzpyida.base import IdaDataBase

class AutoDeleteContext:
    """
    Context manager to lazy removal of temporary tables associated to data frames.
    """
    active_group = threading.local()

    def __init__(self, idadb: IdaDataBase) -> None:
        """
        Creates the context.

        Parameters
        ----------
        idadb : IdaDataBase
            the database connector
        """

        self.idadb = idadb
        self.temp_out_tables = set()

    @staticmethod
    def current():
        """
        Returns an instance of AutoDeleteContext assigned to the current thread or None
        if nothing is assigned.

        Returns
        -------
        AutoDeleteContext
            the current thread AutoDeleteContext or None
        """

        if getattr(AutoDeleteContext.active_group, 'auto_delete_context_st', False) and \
            AutoDeleteContext.active_group.auto_delete_context_st:
            return AutoDeleteContext.active_group.auto_delete_context_st[-1]
        return None

    def add_table_to_delete(self, table_name: str):
        """
        Register a table to be deleted when this context exits.

        Parameters
        ----------
        table_name : str
            the table name
        """
        self.temp_out_tables.add(table_name)

    def __enter__(self):
        if not getattr(AutoDeleteContext.active_group, 'auto_delete_context_st', False):
            AutoDeleteContext.active_group.auto_delete_context_st = []
        AutoDeleteContext.active_group.auto_delete_context_st.append(self)
        return self

    def __exit__(self, ex_type, value, traceback):
        for table_name in self.temp_out_tables:
            if self.idadb.exists_table(table_name):
                self.idadb.drop_table(table_name)
        self.temp_out_tables = set()
        AutoDeleteContext.active_group.auto_delete_context_st.pop()

            


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
import pandas as pd
from nzpyida.base import IdaDataBase

def pytest_addoption(parser):
    """
    Definition of admissible options for the pytest command line
    """
    parser.addoption("--table", default="iris",
        help="Name of the table to test the dataset")
    parser.addoption("--dsn", default="BLUDB",
        help="Data Source Name")
    parser.addoption("--uid", default='',
        help="User ID")
    parser.addoption("--pwd", default='',
        help="Password")
    parser.addoption("--jdbc", default='',
        help="jdbc url string for JDBC connection")
    parser.addoption("--hostname", default='',
        help="hostname for nzpy connection")

@pytest.fixture(scope="session")
def idadb(request):
    """
    DataBase connection fixture, to be used for the whole testing session.
    Hold the main IdaDataBase object. Shall not be closed except by a
    pytest finalizer.
    """
    def fin():
        try:
            idadb.close()
        except:
            pass
    request.addfinalizer(fin)

    hostname = request.config.getoption('--hostname')
    jdbc = request.config.getoption('--jdbc')

    if hostname != '':
        try:
            idadb = IdaDataBase(
                dsn={'database': request.config.getoption('--dsn'), 'host': hostname, 'port': 5480},
                uid=request.config.getoption('--uid'),
                pwd=request.config.getoption('--pwd'),
                autocommit=False)
        except:
            raise
    elif jdbc != '':
        try:
            idadb = IdaDataBase(dsn=jdbc, autocommit=False)
        except:
            raise
    else:
        try:
            idadb = IdaDataBase(dsn=request.config.getoption('--dsn'),
                                        uid=request.config.getoption('--uid'),
                                        pwd=request.config.getoption('--pwd'),
                                        autocommit=False)
        except:
            raise
    return idadb

df_train = pd.DataFrame.from_dict({"ID": range(1000),
                                   "A": [-1, -2, 3, 4, 2, -0.5, 0, 1, -2.1, 1.4]*100,
                                   "B": ['n', 'n', 'p', 'p', 'p', 'n', 'n', 'p', 'n', 'p']*100})
df_test = pd.DataFrame.from_dict({"ID": [0, 1, 2],
                                  "A": [2, 0.001, -2],
                                  "B": ['p', 'p', 'n']})


df_train_reg = pd.DataFrame.from_dict(
    {
        "ID": range(1000),
        "A": range(1, 2001, 2),
        "T": [1000] * 1000,
        "B": range(2, 2001, 2)
    }
)

df_test_reg = pd.DataFrame.from_dict(
    {
        "ID": [0, 1, 2, 3, 4],
        "A": [0, 3, 2222, -1000, 11111],
        "T": [1000] * 5,
        "B": [1, 4, 2223, -999, 11112]
    }
)

df_train_clust = pd.DataFrame.from_dict(
    {
        "ID": range(99),
        "A": list(range(33)) + list(range(100, 133)) + list(range(200, 233))
    }
)

df_test_clust = pd.DataFrame.from_dict(
    {
        "ID": [0, 1, 2, 3],
        "A": [-1, 77, 150, 250]
    }
)

TAB_NAME_TEST = "TAB_NAME1"
TAB_NAME_TRAIN = "TAB_NAME2"

MOD_NAME = "TEST_MOD1"

OUT_TABLE_PRED = "OUT_TABLE_PRED"
OUT_TABLE_CM = "OUT_TABLE_CM"
OUT_TABLE_CV = "OUT_TABLE_CV"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
The discretization process assigns a discrete value to each interval of continuous
attribute a to create a new discrete attribute a'. A discretization algorithm
determines the interval boundaries that are likely to preserve as much useful
information provided by the original attribute as possible. Data set discretization
should preserve the relationship between the class and the discretized attributes
if the data set is to be used for creation of a classification model.
"""
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, materialize_df, make_temp_table_name
from nzpyida.analytics.utils import get_auto_delete_context, call_proc_df_in_out
from nzpyida.analytics.auto_delete_context import AutoDeleteContext


class Discretization:
    """
    Generic class for handling data discretization.
    """

    def __init__(self, idadb: IdaDataBase):
        """
        Creates the discretization class.

        Parameters
        ----------
        idadb : IdaDataBase
            database connector
        """

        self.idadb = idadb
        self.params = {}
        self.proc = ""

    def fit(self, in_df: IdaDataFrame, out_table: str=None) -> IdaDataFrame:
        """
        Create bins limits based on the given data frame.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table with dicretization bins

        Returns
        -------
        IdaDataFrame
            the data frame with discretization bins
        """
        in_columns = ';'.join(in_df.columns)
        params_dict = {
            'incolumn': in_columns,
            'outtabletype': 'table'
        }
        params_dict.update(self.params)

        return call_proc_df_in_out(proc=self.proc, in_df=in_df, params=params_dict,
            out_table=out_table)[0]

    def apply(self, in_df: IdaDataFrame, in_bin_df: IdaDataFrame, keep_org_values: bool=False,
        out_table: str=None) -> IdaDataFrame:
        """
        Apply discretization limits to the given data frame.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        in_bin_df : IdaDataFrame
            the data frame with discretization bins

        keep_org_values : bool, optional
            a flag indicating whether the discretized columns should replace the original
            columns (False) or should be added with another name (True).
            The name of the columns is then prefixed with 'disc_'

        out_table : str, optional
            the output table or view to store the discretized data into

        Returns
        -------
        IdaDataFrame
            the data frame with discerized input data frame
        """
        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        if not isinstance(in_bin_df, IdaDataFrame):
            raise TypeError("Argument in_bin_df should be an IdaDataFrame")

        temp_view_name, need_delete = materialize_df(in_df)
        bin_view_name, bin_view_need_delete = materialize_df(in_bin_df)

        auto_delete_context = None
        if not out_table:
            auto_delete_context = get_auto_delete_context('out_table')
            out_table = make_temp_table_name()

        params = map_to_props({
            'intable': temp_view_name,
            'outtable': out_table,
            'btable': bin_view_name,
            'outtabletype': 'table',
            'replace': not keep_org_values
        })

        try:
            self.idadb.ida_query(f'call NZA..APPLY_DISC(\'{params}\')')
        finally:
            if need_delete:
                self.idadb.drop_view(temp_view_name)
            if bin_view_need_delete:
                self.idadb.drop_view(bin_view_name)

        out_df = IdaDataFrame(self.idadb, out_table)

        if auto_delete_context:
            auto_delete_context.add_table_to_delete(out_table)

        return out_df


class EWDisc(Discretization):
    """
    Discretization with equal width and the given number of bins.
    """

    def __init__(self, idadb: IdaDataBase, bins: int=10):
        """
        Creates the discretization class.

        Parameters
        ----------
        idadb : IdaDataBase
            database connector

        bins : int, optional
            the default number of discretization bins to be calculated
        """

        super().__init__(idadb)
        self.proc = 'EWDISC'
        self.params = {'bins': bins}

def ew_disc(in_df: IdaDataFrame, bins: int=10, keep_org_values: bool=False,
    out_table: str=None) -> IdaDataFrame:
    """
    Discretizes the given data frame with equal width and the given number of bins.
    This is a helper function that creates EWDisc class and then calls its fit() and
    apply() functions, returning the output from the latter.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    bins : int, optional
        the default number of discretization bins to be calculated, by default 10

    keep_org_values : bool, optional
        a flag indicating whether the discretized columns should replace the original
        columns (False) or should be added with another name (True).
        The name of the columns is then prefixed with 'disc_'

    out_table : str, optional
        the output table or view to store the discretized data into

    Returns
    -------
    IdaDataFrame
        the data frame with discerized input data frame
    """

    with AutoDeleteContext(in_df._idadb):
        disc = EWDisc(in_df._idadb, bins=bins)
        bins_df = disc.fit(in_df=in_df, out_table=out_table)
        return disc.apply(in_df=in_df, in_bin_df=bins_df, keep_org_values=keep_org_values)

class EFDisc(Discretization):
    """
    Discretization with equal frequency of data.
    """

    def __init__(self, idadb: IdaDataBase, bins: int=10, bin_precision: float=0.1):
        """
        Creates the discretization class.

        Parameters
        ----------
        idadb : IdaDataBase
            database connector

        bins : int, optional
            the default number of discretization bins to be calculated

        bin_precision : float, optional
            the precision allowed for considering an even distribution of data records in
            the calculated discretization bins. The number of data records in each bin
            must be within [iw-<binprec>*iw,iw+<binprec>*iw] where iw is the size of the
            input table divided by the number of requested discretization bin limits.
        """

        super().__init__(idadb)
        self.proc = 'EFDISC'
        self.params = {'bins': bins, 'binprec': bin_precision}

def ef_disc(in_df: IdaDataFrame, bins: int=10, bin_precision: float=0.1,
    keep_org_values: bool=False, out_table: str=None):
    """
    Discretizes the given data frame with equal frequency of data.
    This is a helper function that creates EFDisc class and then calls its fit() and
    apply() functions, returning the output from the latter.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    bins : int, optional
        the default number of discretization bins to be calculated, by default 10

    bin_precision : float, optional
        the precision allowed for considering an even distribution of data records in
        the calculated discretization bins. The number of data records in each bin
        must be within [iw-<binprec>*iw,iw+<binprec>*iw] where iw is the size of the
        input table divided by the number of requested discretization bin limits.

    keep_org_values : bool, optional
        a flag indicating whether the discretized columns should replace the original
        columns (False) or should be added with another name (True).
        The name of the columns is then prefixed with 'disc_'

    out_table : str, optional
        the output table or view to store the discretized data into

    Returns
    -------
    IdaDataFrame
        the data frame with discerized input data frame
    """
    with AutoDeleteContext(in_df._idadb):
        disc = EFDisc(in_df._idadb, bins=bins, bin_precision=bin_precision)
        bins_df = disc.fit(in_df=in_df, out_table=out_table)
        return disc.apply(in_df=in_df, in_bin_df=bins_df, keep_org_values=keep_org_values)

class EMDisc(Discretization):
    """
    Discretization based on minimizing entropy of the data in the target column.
    """

    def __init__(self, idadb: IdaDataBase, target: str):
        """
        Creates the discretization class.

        Parameters
        ----------
        idadb : IdaDataBase
            database connector

        target : str
            the input table column containing a class label
        """

        super().__init__(idadb)
        self.proc = 'EMDISC'
        self.params = {'target': target}

def em_disc(in_df: IdaDataFrame, target: str, keep_org_values: bool=False, out_table: str=None):
    """
    Discretizes the given data frame based on minimizing entropy of the data 
    in the target column.
    This is a helper function that creates EMDisc class and then calls its fit() and
    apply() functions, returning the output from the latter.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    target : str
        the input table column containing a class label

    keep_org_values : bool, optional
        a flag indicating whether the discretized columns should replace the original
        columns (False) or should be added with another name (True).
        The name of the columns is then prefixed with 'disc_'

    out_table : str, optional
        the output table or view to store the discretized data into

    Returns
    -------
    IdaDataFrame
        the data frame with discerized input data frame
    """
    with AutoDeleteContext(in_df._idadb):
        disc = EMDisc(in_df._idadb, target=target)
        bins_df = disc.fit(in_df=in_df, out_table=out_table)
        return disc.apply(in_df=in_df, in_bin_df=bins_df, keep_org_values=keep_org_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
Linear regression is a simple but very useful and commonly applied approach
to the regression task, even though it only performs direct modeling of
linear relationships. It is the thing that limits its applicability, a linear
model representation, that makes it fast, efficient, and easy to use (compared
to more refined regression algorithms).
"""
from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase

from nzpyida.analytics.predictive.regression import Regression


class LinearRegression(Regression):
    """
    Linear regression predictive model.
    """

    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the regressor class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.fit_proc = 'LINEAR_REGRESSION'
        self.predict_proc = 'PREDICT_LINEAR_REGRESSION'


    def fit(self, in_df: IdaDataFrame, target_column: str, id_column: str=None,
        in_columns: List[str]=None, nominal_colums: str=None, col_def_type: str=None,
        col_def_role: str=None, col_properties_table: str=None, use_svd_solver: bool=False,
        intercept: bool=True, calculate_diagnostics: bool=False):
        """
        Creates a linear regression model based on provided data and store it in a database.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame

        target_column : str
            the input table column representing the prediction target, definition of multitargets
            can be processed by 'incolumn' parameter and column properties.

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        nominal_colums : str, optional
            the input table nominal columns, if any, separated by a semi-colon (;).
            Parameter 'nominalCols' is deprecated please use 'incolumn' intead.

        in_columns : List[str], optional
            the list of input table columns with special properties.
            Each column is followed by one or several of the following properties:
                its type: ':nom' (for nominal), ':cont' (for continuous).
                Per default, all numerical types are con-tinuous, other types are nominal.
                its role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same
            as ':colweight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns. Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input table column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight' is
            unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsup-ported,
            i.e. same as '1')

        use_svd_solver : bool, optional
            a flag indicating whether Singular Value Decomposition and matrix multiplication
            should be used for solving the matrix equation

        intercept : bool, optional
            flag indicating whether the model is built with or without an intercept value.
            The default has changed to true.

        calculate_diagnostics : bool, optional
            a flag indicating whether diagnostics information should be displayed
        """

        params = {
            'id': id_column,
            'target': target_column,
            'nominalCols': nominal_colums,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
            'useSVDSolver': use_svd_solver,
            'intercept': intercept,
            'calculateDiagnostics': calculate_diagnostics
        }

        self._fit(in_df=in_df, params=params)



from nzpyida.ae.install import NZInstall
nzpy_dsn = {
        "database":"housing",
         "port" :5480,
        "host": "9.30.250.118",
        "securityLevel":3,
        "logLevel":0
       }
import pandas as pd
from nzpyida import IdaDataBase, IdaDataFrame
from nzpyida.ae import NZFunTApply
idadb = IdaDataBase(nzpy_dsn, uid="admin", pwd="password", verbose=True)

#idadb.ida_query("create view houseprice_temp as select * from houseprice")

#idadb.ida_query("alter table houseprice add column date_str varchar(1000)")
#idadb.ida_query("update  houseprice set date_str = cast(ID as varchar(1000))")
#idadf = IdaDataFrame(idadb,"HOUSEPRICE")
idadf = IdaDataFrame(idadb,"HOUSEPRICE_TEMP")

pd.set_option('display.max_rows', None)
pd.set_option('display.max_columns', None)
print(idadf.head())
print(idadf.dtypes)
#package_name = 'sklearn'
#nzinstall = NZInstall(idadb, package_name)
#result = nzinstall.getResultCode()
#print(result)
code_house_prediction_model = """def house_prediction_model(self,df):
            import numpy as np
            import pandas as pd
            import nzaeCppWrapper

            import sklearn

            from sklearn.model_selection import cross_val_score
            from sklearn.model_selection import train_test_split
            from sklearn.linear_model import LinearRegression
            # from sklearn.ensemble import GradientBoostingRegression
            # from xgboost import XGBRegressor
            from sklearn.metrics import r2_score,mean_squared_error
            from datetime import datetime

            # Getting the current date and time
            #dt = datetime.now()

            # getting the timestamp
            #ts = datetime.timestamp(dt)



            data = df.copy()

            #id1 = data['ID'][0]
            #id = id1/100000
            #data['test_field'] = str(ts)
            
           

            


            
            

            


            

            


           

            self.output(id,0.85, 'hello')











"""

output_signature = {'ID': 'float', 'accuracy': 'double', 'date':'str'}

nz_fun_tapply = NZFunTApply(df=idadf, code_str=code_house_prediction_model, fun_name="house_prediction_model",
                            parallel=False,  output_signature=output_signature)

print(nz_fun_tapply.get_result())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.analytics.predictive.timeseries import TimeSeries
from nzpyida.base import IdaDataBase
from nzpyida.frame import IdaDataFrame
from nzpyida.analytics.model_manager import ModelManager
import pytest
from nzpyida.analytics.tests.conftest import MOD_NAME, TAB_NAME_TRAIN, OUT_TABLE_PRED
import pandas as pd
from math import sin


@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture
def clean_up(idadb, mm):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)


@pytest.fixture
def idf_train(idadb: IdaDataBase):
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)

    time_series = [sin(x)+x for x in range(200)]
    df = pd.DataFrame.from_dict({
        "TIME": range(200),
        "VALUE": time_series
    })
    yield idadb.as_idadataframe(df, TAB_NAME_TRAIN)

    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)


def test_timeseries(idadb: IdaDataBase, mm: ModelManager, idf_train: IdaDataFrame, clean_up):
    model = TimeSeries(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME) 

    outtab = model.fit_predict(idf_train, time_column="TIME", target_column="VALUE", out_table=OUT_TABLE_PRED, 
                       forecast_horizon='399')

    assert mm.model_exists(MOD_NAME) 
    assert outtab
    assert len(outtab) == 200
    assert round(outtab.head(10).iloc[-1]["VALUE"]) == round(sin(210)+210)
    assert round(outtab.tail().iloc[-1]["VALUE"]) == round(sin(399)+399)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
from typing import Dict, Tuple, Any
from time import time
import pandas as pd
import random
import re
from nzpyida.frame import IdaDataFrame
from nzpyida.analytics.auto_delete_context import AutoDeleteContext


def map_to_props(data: Dict[str, Any]) -> str:
    """
    Maps the given dictionary of attributes into coma separated list of key/value
    pairs. Skips these pairs with value None.

    Parameters
    ----------
    data : dict
        the input data

    Returns
    -------
    str
        the output string
    """

    ret = []
    for k, v in data.items():
        if isinstance(v, list):
            if len(v):
                ret.append(f"{k}={';'.join(map(str, v))}")
        elif v is not None:
            ret.append(f"{k}={v}")
    return ",".join(ret)

def materialize_df(df: IdaDataFrame) -> Tuple[str, bool]:
    """
    Creates a view associated to the given data frame in the database.
    If the data frame requires no view, simply returns a table assiciated with it.
    The second items of the returned tuple shows if what is returned is a temporaty view
    (that should be dropped by the caller) or exisintg table.

    Parameters
    ----------
    df : IdaDataFrame
        the input data frame

    Returns
    -------
    str
        the name of db object (view or table) available representing the data frame

    bool
        should the object be dropped by the caller
    """

    if df.internal_state.views:
        temp_view_name = make_temp_table_name()
        query = f'CREATE VIEW "{temp_view_name}" AS ({df.internal_state.get_state()})'
        df.ida_query(query, autocommit = True)
        return temp_view_name, True
    else:
        return df.tablename, False

def make_temp_table_name(prefix: str='DATA_FRAME_') -> str:
    """
    Generate temp table name.

    Parameters
    ----------
    prefix : str, optional
        name prefix

    Returns
    -------
    str
        generated table name with the given prefix
    """

    return f'{prefix}{random.randint(0, 100000)}_{int(time())}'

def get_auto_delete_context(out_table_attr_name: str) -> AutoDeleteContext:
    """
    Returns an auto delete context manager assigned to the current thread.
    If nothing is assigned, raise an exception with appropriate error message.
    The 'out_table_attr_name' is included in that message.

    Parameters
    ----------
    out_table_attr_name : str
        the name of function parameter with the output table name (used for
        a message in the exception only)

    Returns
    -------
    AutoDeleteContext
        current thread AutoDeleteContext

    Raises
    ------
    RuntimeError
        If there is no AutoDeleteContext attached to the current thread.
    """

    if AutoDeleteContext.current() is None:
        raise RuntimeError('This code needs to run inside of AutoDeleteContext context manager or '
        f'you need to set an output table name in {out_table_attr_name} function attribute')
    return AutoDeleteContext.current()

def call_proc_df_in_out(proc: str, in_df: IdaDataFrame, params: dict,
    out_table: str=None, copy_indexer=False) -> Tuple[IdaDataFrame, str]:
    """
    Generic function for data processing.
    """
    if not isinstance(in_df, IdaDataFrame):
        raise TypeError("Argument in_df should be an IdaDataFrame")

    temp_view_name, need_delete = materialize_df(in_df)

    auto_delete_context = None
    if not out_table:
        auto_delete_context = get_auto_delete_context('out_table')
        out_table = make_temp_table_name()

    params['intable'] = temp_view_name
    params['outtable'] = out_table

    params_s = map_to_props(params)

    try:
        out_query = in_df.ida_query(f'call NZA..{proc}(\'{params_s}\')')
    finally:
        if need_delete:
            in_df._idadb.drop_view(temp_view_name)

    if not in_df._idadb.exists_table_or_view(out_table):
        # stored procedure call was successful by did not produce a table
        return None, out_query

    if auto_delete_context:
        auto_delete_context.add_table_to_delete(out_table)

    out_df = IdaDataFrame(in_df._idadb, out_table)
    if copy_indexer and in_df.indexer:
        out_df.indexer = in_df.indexer
    return out_df, out_query

def out_str_to_df(out_str: str):
    clean_str = re.sub('[^a-zA-Z0-9" ]', '', out_str)
    clean_str_split =clean_str.split(' ')
    out_dict = {}
    for i in range(0, len(clean_str_split), 2):
        try:
            value = int(clean_str_split[i+1])
        except ValueError:
            value = clean_str_split[i+1]
        out_dict[clean_str_split[i].upper()] = [value]
    return  pd.DataFrame.from_dict(out_dict)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
from .predictive.decision_trees import DecisionTreeClassifier
from .predictive.kmeans import KMeans
from .predictive.knn import KNeighborsClassifier
from .predictive.linear_regression import LinearRegression
from .predictive.naive_bayes import NaiveBayesClassifier
from .predictive.association_rules import ARule
from .predictive.bisecting_kmeans import BisectingKMeans
from .predictive.regression_trees import DecisionTreeRegressor
from .predictive.two_step_clustering import TwoStepClustering
from .predictive.timeseries import TimeSeries
from .predictive.bayesian_networks import TreeBayesNetwork, BinaryTreeBayesNetwork, \
MultiTreeBayesNetwork, TreeBayesNetwork1G, TreeBayesNetwork1G2P, TreeBayesNetwork2G, \
TreeAgumentedNetwork
from .predictive.glm import BernoulliRegressor, BinomialRegressor, \
    NegativeBinomialRegressor, GaussianRegressor, GammaRegressor, \
    PoissonRegressor, WaldRegressor
from .exploration.distribution import bitable, moments, histogram, outliers
from .exploration.distribution import quantile, unitable
from .transform.discretization import EFDisc, EMDisc, EWDisc
from .transform.discretization import ef_disc, em_disc, ew_disc
from .transform.preparation import std_norm, impute_data, random_sample, train_test_split
from .model_manager import ModelManager
from .auto_delete_context import AutoDeleteContext
from .exploration.relation_identification import corr, cov, covariance_matrix, \
spearman_corr, mutual_info, chisq, t_me_test, t_ls_test, t_pmd_test, t_umd_test, \
mww_test, wilcoxon_test, canonical_corr, anova_crd_test, anova_rbd_test, \
manova_one_way_test, manova_two_way_test




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.analytics.predictive.decision_trees import DecisionTreeClassifier
from nzpyida.analytics.tests.conftest import MOD_NAME, OUT_TABLE_PRED, OUT_TABLE_CM, TAB_NAME_TEST, \
    TAB_NAME_TRAIN, df_test, df_train
import pandas as pd
import pytest

TAB_NAME_TEST = "TAB_NAME1"
TAB_NAME_TRAIN = "TAB_NAME2"
MOD_NAME = "MOD_NAME1"

@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture(scope='module')
def clear_up(idadb: IdaDataBase, mm: ModelManager):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CM):
        idadb.drop_table(OUT_TABLE_CM)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CM):
        idadb.drop_table(OUT_TABLE_CM)

def test_decision_trees(idadb: IdaDataBase, mm: ModelManager, clear_up):
    idf_train = idadb.as_idadataframe(df_train, tablename=TAB_NAME_TRAIN, clear_existing=True)
    idf_test = idadb.as_idadataframe(df_test, tablename=TAB_NAME_TEST, clear_existing=True)
    assert idf_train
    assert idf_test

    model = DecisionTreeClassifier(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME)
    
    model.fit(idf_train, id_column="ID", target_column="B", eval_measure='gini', min_improve=0, min_split=200)
    assert mm.model_exists(MOD_NAME)

    pred = model.predict(idf_test, id_column="ID", out_table=OUT_TABLE_PRED)
    assert pred
    assert all(pred.columns == ['ID', 'CLASS'])
    assert list(pred.as_dataframe()['CLASS'].values) in (['p', 'n', 'n'], ['p', 'p', 'n'])

    score = model.score(idf_test, id_column="ID", target_column="B")

    assert score
    assert score >= 0.66

    cm, acc, wacc = model.conf_matrix(idf_test, id_column='ID', target_column='B', out_matrix_table=OUT_TABLE_CM)
    assert all([cm, acc, wacc])
    assert all(cm.columns == ['REAL', 'PREDICTION', 'CNT'])
    assert len(cm) >= 2
    assert sum(cm.as_dataframe()["CNT"].values) == 3
    assert wacc >= 0.75
    assert acc >= 0.66




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.analytics.predictive.two_step_clustering import TwoStepClustering

from nzpyida.analytics.tests.conftest import MOD_NAME, OUT_TABLE_PRED, TAB_NAME_TEST, \
    TAB_NAME_TRAIN, df_train_clust, df_test_clust
import pandas as pd
import pytest

OUT_TABLE_CLUST = "OUT_TABLE_CLUST"

@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture(scope='module')
def clear_up(idadb: IdaDataBase, mm: ModelManager):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CLUST):
        idadb.drop_table(OUT_TABLE_CLUST)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CLUST):
        idadb.drop_table(OUT_TABLE_CLUST)

def test_bisecting_kmeans(idadb: IdaDataBase, mm: ModelManager, clear_up):
    idf_train = idadb.as_idadataframe(df_train_clust, tablename=TAB_NAME_TRAIN, clear_existing=True, indexer='ID')
    idf_test = idadb.as_idadataframe(df_test_clust, tablename=TAB_NAME_TEST, clear_existing=True, indexer='ID')
    assert idf_train
    assert idf_test

    model = TwoStepClustering(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME)
    
    model.fit(idf_train, distance='euclidean', max_k=5, bins=10, node_capacity=10, leaf_capacity=10,
              max_leaves=100, rand_seed=4321, out_table=OUT_TABLE_CLUST)
    assert mm.model_exists(MOD_NAME)

    pred = model.predict(idf_test, out_table=OUT_TABLE_PRED)
    assert pred
    assert all(pred.columns == ['ID', 'CLUSTER_ID', 'DISTANCE'])
    assert all(list(pred.head()['CLUSTER_ID'].values))

    score = model.score(idf_test, target_column="A")

    assert score



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
"""
This module contains function that can be used to prepare an input data
frame for machine learning.
"""
from typing import List, Tuple
from nzpyida.frame import IdaDataFrame
from nzpyida.analytics.utils import materialize_df, make_temp_table_name, \
    get_auto_delete_context, call_proc_df_in_out, map_to_props


def std_norm(in_df: IdaDataFrame, in_column: List[str], id_column: str = None,
             by_column: str = None, out_table: str = None) -> IdaDataFrame:
    """
    Standardization and normalization transformations use the original continuous
    attribute a to generate a new continuous attribute a ' that has a different range
    or distribution than the original attribute. Common transformations modify the
    range to fit the [-1,1 ] interval (normalization) or modify the distribution to
    have a mean of 0 and a standard deviation of 1 (standardization).

    This function normalize and stardardize columns of the input data frame and returns
    that in a new data frame.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : List[str]
        the list of input table columns to consider.
        Each column name may be followed by :L to leave it unchanged, by :S to standardize
        its values, by :N to normalize its values or by :U to make it of unit length.
        Additionally, two columns may be indicated, separated by a slash (/), followed
        by :C to make the columns be a row unit vector or by :V to divide the column
        values by the length of the longest row vector.

    id_column : str, optional
        the input table column identifying a unique instance id - if skipped, 
        the input data frame indexer must be set and will be used as an instance id

    by_column : str, optional
        the input table column which splits the data into groups for which the operation
        is to be performed

    out_table : str, optional
        the output table with the modified data

    Returns
    -------
    IdaDataFrame
        the data frame with requested transformations
    """
    if not id_column:
        if in_df.indexer:
            id_column = in_df.indexer
        else:
            raise TypeError('Missing id column - either use id_column attribute or set '
                            'indexer column in the input data frame')

    params = {
        'id': id_column,
        'incolumn': in_column,
        'by': by_column
    }
    return call_proc_df_in_out(proc='STD_NORM', in_df=in_df, params=params,
                               out_table=out_table, copy_indexer=True)[0]


def impute_data(in_df: IdaDataFrame, in_column: str = None, method: str = None,
                numeric_value: float = -1, nominal_value: str = 'missing', out_table: str = None) -> IdaDataFrame:
    """
    Many analytic algorithms require that the data set has no missing attribute values.
    However, real-world data sets frequently suffer from missing attribute values.
    Missing value imputation provides usable attribute values in place of the missing values,
    allowing the algorithms to run.

    This function replaces missing values in the input data frame and returns the result
    in a new data frame.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    in_column : str, optional
        the input table column where missing values have to be replaced.
        If not specified, all input data columns are considered.

    method : str, optional
        the data imputation method. Allowed values are: mean, median, freq (most frequent value),
        replace. If not specified, the method is median for the numeric columns and freq for
        the nominal columns. The methods mean and median cannot be used with nominal columns.

    numeric_value : float, optional
        the numeric replacement value when method=replace

    nominal_value : str, optional
        the nominal replacement value when method=replace

    out_table : str, optional
        the output table with the modified data

    Returns
    -------
    IdaDataFrame
        the data frame with requested transformations
    """

    params = {
        'incolumn': in_column,
        'method': method,
        'numericvalue': numeric_value,
        'nominalvalue': nominal_value
    }
    return call_proc_df_in_out(proc='IMPUTE_DATA', in_df=in_df, params=params,
                               out_table=out_table, copy_indexer=True)[0]


def random_sample(in_df: IdaDataFrame, size: int = None, fraction: float = None, by_column: str = None,
                  out_signature: str = None, rand_seed: int = None, out_table: str = None) -> IdaDataFrame:
    """
    Random sampling procedures are a vital component of many analytical systems. They can
    be used to select a test sample and a training sample for a model building process
    (machine learning). They can also be used to get a smaller sample of the training
    set, which you may do because of learning algorithm complexity considerations.
    In both cases, you would sample without replacement.

    Another application of sampling is the learning methods based on bootstrapping.
    This requires many independent samples from the same data, which are preferentially
    applied if the available data sets are small or for other reasons where the sample
    independence is vital. Samples with replacement are usually drawn in this case.

    In application, sampling is used for promotion campaigns, for example when you want
    only a representative set of customers to be subjects of an action.
    In all cases, whether for use with scientific methods or business practices, uniform
    sampling is important.

    This function creates a random sample of a data frame a fixed size or a fixed
    probability and returns the result in a new data frame.

    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    size : int, optional
        the number of rows in the sample (alias of size). If specified,
        the parameter <fraction> must not be specified. Only one of both parameters <num>
        and <size> must be specified.

    fraction : float, optional
        the probability of each row to be in the sample. If specified, the parameters <num>
        and <size> must not be specified. Otherwise, one of both parameters <num> or <size>
        must be specified.

    by_column : str, optional
        the column used to stratify the input table. If indicated, stratified sampling is
        done: it ensures that each value of the column is represented in the sample in
        about the same percentage as in the original input table.

    out_signature : str, optional
        the input table columns to keep in the sample, separated by a semi-colon (;).
        If not specified, all columns are kept in the output table.

    rand_seed : int, optional
        the seed of the random function

    out_table : str, optional
        the output table with the modified data

    Returns
    -------
    IdaDataFrame
        the data frame with requested transformations
    """

    params = {
        'size': size,
        'fraction': fraction,
        'by': by_column,
        'outsignature': out_signature,
        'randseed': rand_seed
    }
    return call_proc_df_in_out(proc='RANDOM_SAMPLE', in_df=in_df, params=params,
                               out_table=out_table, copy_indexer=True)[0]


def train_test_split(in_df: IdaDataFrame, out_table_train: str=None, out_table_test: str=None,
                     id_column: str = None, fraction: float = 0.5, 
                     rand_seed: float = None) -> Tuple[IdaDataFrame, IdaDataFrame]:
    """
    Parameters
    ----------
    in_df : IdaDataFrame
        the input data frame

    out_table_train : str, optional
        the name of output dataframe that will contain the given fraction of the input records

    out_table_test : str, optional
        the name of output dataframe that will contain the rest (1-<fraction>) of the input records

    id_column: str, optional
        the input dataframe column identifying a unique instance id

    fraction : float, optional
        the fraction of the data to that goes to the training dataframe

    rand_seed : int, optional
        the seed of the random function

    Returns
    -------
    IdaDataFrame
        the data frame with train data
    
    IdaDataFrame
        the data frame with test data
    """
    
    if not isinstance(in_df, IdaDataFrame):
        raise TypeError("Argument in_df should be an IdaDataFrame")
    
    if not id_column:
        if not in_df.indexer:
            raise ValueError("If dataframe has no indexer 'id_column' has to be provided")
        else:
            id_column = in_df.indexer
    temp_view_name, need_delete = materialize_df(in_df)
    
    auto_delete_context = None
    if not out_table_train:
        auto_delete_context = get_auto_delete_context('out_table_train')
        out_table_train = make_temp_table_name()

    if not out_table_test:
        auto_delete_context = get_auto_delete_context('out_table_test')
        out_table_test = make_temp_table_name()

    params = {
        'intable': temp_view_name,
        'traintable': out_table_train,
        'testtable': out_table_test,
        'id': id_column,
        'fraction': fraction
    }

    if isinstance(rand_seed, float):
        params['seed'] = rand_seed

    params_s = map_to_props(params)

    try:
        in_df.ida_query(f'call NZA..SPLIT_DATA(\'{params_s}\')')
    finally:
        if need_delete:
            in_df._idadb.drop_view(temp_view_name)

    if not in_df._idadb.exists_table_or_view(out_table_train) or \
       not in_df._idadb.exists_table_or_view(out_table_test):
        # stored procedure call was successful by did not produce a table
        return None, None

    if auto_delete_context:
        auto_delete_context.add_table_to_delete(out_table_train)
        auto_delete_context.add_table_to_delete(out_table_test)

    out_df_train = IdaDataFrame(in_df._idadb, out_table_train)
    out_df_test = IdaDataFrame(in_df._idadb, out_table_test)
    
    if in_df.indexer:
        out_df_train.indexer = in_df.indexer
        out_df_test.indexer = in_df.indexer
    return out_df_train, out_df_test


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.analytics.utils import call_proc_df_in_out, make_temp_table_name, out_str_to_df
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from typing import List
import pandas as pd
from nzpyida.analytics.predictive.regression import Regression


class GLM(Regression):
    """ 
    General Linear Regression model
    """
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates GLM object

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """
        super().__init__(idadb, model_name)
        self.fit_proc = "GLM"
        self.predict_proc = "PREDICT_GLM"
        self.has_print_proc = True
        self.target_column_in_output = 'PRED'

    
    def fit(self, in_df: IdaDataFrame, target_column: str, id_column: str=None, in_columns: List[str]=None,  
            intercept: bool=True, interaction: str='', family_param: float=-1, link: str='logit', 
            link_param: float=1, max_iter: int=20, epsilon: float=1e-3, tolerance: float=1e-7, 
            method: str='irls', trials: str='', debug: bool=False, col_def_type: str=None, 
            col_def_role: str=None, col_properties_table: str=None):
        """
        in_df : IdaDataFrame
            the input data frame
        
        target_column : str
            the input dataframe column to predict a value for. Only numeric type of target column is
            accepted
        
        id_column : str, optional
            the input datafrme column identifying a unique instance id
        
        incolumn : str, optional
            the list of input dataframe columns with special properties, separated. Each column is 
            followed by one or several of the following properties:
            - its type: ':nom' (for nominal), ':cont' (for continuous). Per default, all numerical 
              types are con-tinuous, other types are nominal.
            - its role: ':id', ':target', ':input', ':ignore', ':objweight'.
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same as 
            ':colweight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties
        
        intercept: bool, optional
            flag indicating whether the model is built with or without an intercept value
        
        interaction: str, optional
            the definition of the allowed interactions between input columns. The interaction 
            is a list of factors separated by a semicolon (;). A factor is a list of variables 
            separated by a star (*). A variable is a column name of the input table. Continuous 
            variables can be followed by a caret (^) and a numeric value, in this case the given 
            power of values of this column is meant. Nominal variables can be followed by a sign 
            equal (=) and a value, so that only the given value of the variable is allowed to 
            interact with the other variables of this factor. If no value is indicated after 
            a nominal variable, all distinct val- ues interact independantly with the other 
            variables of the factor. By default, all input columns are considered independent 
            and do not interact with each other
        
        family_param: float, optional
            additional parameter used for some distributions. IF family_param='quasi' then 
            quasi-likelihood in case of Poisson and Binomial distributions is optimized. 
            IF family_param=-1 (or is omitted then mentioned distribution parameter is estimated 
            from data. IF family_param is given explicit then should by > 0
        
        link: str, optional
            the type of the link function. Allowed values are: canbinom, cangeom, cannegbinom, 
            cauchit, clog, cloglog, gaussit, identity, inverse, invnegative, invsquare, log, logit, 
            loglog, oddspower, power, probit, sqrt
        
        link_param: float, optional
            an additional parameter used for some links like: cannegbinom, oddspower, power. 
            The range of value depends on the used link function
        
        max_iter: int, optional
            the maximum number of iterations
        
        epsilon: float, optional
            the maximum (relative) error used as stopping criteria
        
        tolerance: float, optional
            the tolerance for the linear equation solver when to consider a value to be equal to zero
        
        method: str, optional
            the method used to calculate a GLM model. Allowed values are: irls, psgd
        
        trials: str, optional
            the input table column containing the number of trials for the binominal distribution. 
            This parameter must be specified when family=binomial. 
            This parametrs is ignored for other distributions
        
        debug: str, optional
            flag indicating to display debug information
        
        col_def_type: str, optional
            default type of the input dataframe columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous, 
            other columns nominal
        
        col_def_role: str, optional
            default role of the input dataframe columns. Allowed values are 'input' and 'ignore'. 
            If the parameter is undefined, all columns are considered 'input' columns

        col_properties_table: str, optional
            the input table where column properties for the input dataframe columns are stored. 
            The format of this table is the output format of stored procedure nza..COLUMN_PROPER-TIES().
            If the parameter is undefined, the input table column properties will be detected automatically.
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported, i.e. same as '1')
        """
        params = {
            'family': self.family,
            'target': target_column,
            'id': id_column,
            'incolumn': in_columns,
            'coldefrole': col_def_role,
            'coldeftype': col_def_type,
            'colPropertiesTable': col_properties_table,
            'intercept': intercept,
            'family_param': family_param,
            'link': link,
            'link_param': link_param,
            'maxit': max_iter,
            'eps': epsilon,
            'tol': tolerance,
            'method': method,
            'debug': debug
        }
        if interaction:
            params['interaction'] = interaction
        if trials:
            params['trials'] = trials

        return self._fit(in_df=in_df, params=params)

    def predict(self, in_df: IdaDataFrame, out_table: str=None, id_column: str=None, 
                debug: bool=False):
        """
        in_df : IdaDataFrame
            the input data frame
        
        out_table : str, optional
            the output table where the predictions will be stored
        
        id_column : str, optional
            the input data frame column identifying a unique instance
        
        debug : bool, optional
            flag indicating to display debug information
        
        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """
        params = {
            'id': id_column,
            'debug': debug
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)

class BernoulliRegressor(GLM):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        super().__init__(idadb, model_name)
        self.family = 'bernoulli'


class BinomialRegressor(GLM):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        super().__init__(idadb, model_name)
        self.family = 'binomial'

class PoissonRegressor(GLM):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        super().__init__(idadb, model_name)
        self.family = 'poisson'

class NegativeBinomialRegressor(GLM):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        super().__init__(idadb, model_name)
        self.family = 'negativebinomial,'

class GaussianRegressor(GLM):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        super().__init__(idadb, model_name)
        self.family = 'gaussian'

class WaldRegressor(GLM):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        super().__init__(idadb, model_name)
        self.family = 'wald'

class GammaRegressor(GLM):
    def __init__(self, idadb: IdaDataBase, model_name: str):
        super().__init__(idadb, model_name)
        self.family = 'gamma'




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
from nzpyida.base import IdaDataBase
from nzpyida.analytics.auto_delete_context import AutoDeleteContext
from nzpyida.analytics.tests.conftest import df_train, TAB_NAME_TRAIN
from nzpyida.analytics.transform.discretization import EFDisc, EMDisc, EWDisc



@pytest.fixture(scope='module')
def idf(idadb: IdaDataBase):
    idf = idadb.as_idadataframe(df_train, tablename=TAB_NAME_TRAIN, clear_existing=True, indexer='ID')[:10]['A']
    yield idf
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)


def test_ewdisc(idadb: IdaDataBase, idf):
    with AutoDeleteContext(idadb):
        ewdisc = EWDisc(idadb, bins=5)
        bin_df = ewdisc.fit(idf)
        assert bin_df
        assert len(bin_df) == 4
        assert all(bin_df.columns == ['COLNAME', 'BREAK'])

        ew_df = ewdisc.apply(idf, in_bin_df=bin_df, keep_org_values=True)
        assert ew_df
        assert len(ew_df) == len(idf)
        assert all(ew_df.columns == ['A', 'DISC_A'])
        assert set(ew_df['DISC_A'].head(10).values) == {'1', '2', '3', '4', '5'}


def test_efdisc(idadb: IdaDataBase, idf):
    with AutoDeleteContext(idadb):
        efdisc = EFDisc(idadb, bins=2)
        bin_df = efdisc.fit(idf)
        assert bin_df
        assert len(bin_df) == 1
        assert all(bin_df.columns == ['COLNAME', 'BREAK'])

        ef_df = efdisc.apply(idf, in_bin_df=bin_df, keep_org_values=True)
        assert ef_df
        assert len(ef_df) == len(idf)
        assert all(ef_df.columns == ['A', 'DISC_A'])
        assert set(ef_df['DISC_A'].head(10).values) == {'1', '2'}


def test_emdisc(idadb: IdaDataBase, idf):
    with AutoDeleteContext(idadb):
        emdisc = EMDisc(idadb, target='A')
        bin_df = emdisc.fit(idf)
        assert bin_df
        assert all(bin_df.columns == ['COLNAME', 'BREAK'])

        em_df = emdisc.apply(idf, in_bin_df=bin_df, keep_org_values=True)
        assert em_df
        assert len(em_df) == len(idf)
        assert all(em_df.columns == ['A', 'DISC_A'])
        assert len(set(em_df['DISC_A'].head(10).values)) < len(idf)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
This module contains a class that is the base for all regression algorithms.
"""
from typing import Dict
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, materialize_df, make_temp_table_name
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling


class Regression(PredictiveModeling):
    """
    Base class for regression algorithms.
    """

    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the regressor class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.score_proc = 'MSE'
        self.id_column_in_output = 'ID'

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame to predict

        out_table : str, optional
            the output table where the predictions will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """

        params = {
            'id': id_column
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)

    def score(self, in_df: IdaDataFrame, target_column: str,
        id_column: str=None) -> float:
        """
        Scores the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        Returns
        -------
        float
            the model score
        """

        params = {
            'id': id_column
        }

        return self._score(in_df=in_df, predict_params=params, target_column=target_column)

    def score_all(self, in_df: IdaDataFrame, target_column: str,
        id_column: str=None) -> Dict[str, float]:
        """
        Scores the model using MSE, MAE, RSE and RAE. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        Returns
        -------
        dict
            the model scores in a dictionary with MSE, MAE, RSE and RAE as keys
        """
        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        if not id_column:
            if in_df.indexer:
                id_column = in_df.indexer
            else:
                raise TypeError('Missing id column - either use id_column attribute or set '
                    'indexer column in the input data frame')

        out_table = make_temp_table_name()

        pred_view_needs_delete, true_view_needs_delete = False, False
        try:
            pred_df = self.predict(in_df=in_df, out_table=out_table, id_column=id_column)

            pred_view, pred_view_needs_delete = materialize_df(pred_df)
            true_view, true_view_needs_delete = materialize_df(in_df)

            params = map_to_props({
                'pred_table': pred_view,
                'true_table': true_view,
                'pred_id': id_column if self.id_column_in_output is None
                    else self.id_column_in_output,
                'true_id': id_column,
                'pred_column': target_column if self.target_column_in_output is None
                    else self.target_column_in_output,
                'true_column': target_column
            })

            res1 = pred_df.ida_query(f'call NZA..MSE(\'{params}\')')
            res2 = pred_df.ida_query(f'call NZA..MAE(\'{params}\')')
            res3 = pred_df.ida_query(f'call NZA..RSE(\'{params}\')')
            res4 = pred_df.ida_query(f'call NZA..RAE(\'{params}\')')
            res_dict = {
                "MSE": res1[0],
                "MAE": res2[0],
                "RSE": res3[0],
                "RAE": res4[0]
            }
            return res_dict
        finally:
            self.idadb.drop_table(out_table)
            if pred_view_needs_delete:
                self.idadb.drop_view(pred_view)
            if true_view_needs_delete:
                self.idadb.drop_view(true_view)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
The k-means algorithm is the most widely-used clustering algorithm that uses
an explicit distance measure to partition the data set into clusters.
The main concept behind the k-means algorithm is to represent each cluster
by the vector of mean attribute values of all training instances assigned
to that cluster, called the cluster’s center. There are direct consequences
of such a cluster representation:

- the algorithm handles continuous attributes only, although workarounds
for discrete attributes are possible

- both the cluster formation and cluster modeling processes can be performed
in a computationally efficient way by applying the specified distance
function to match instances against cluster centers
"""

from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, make_temp_table_name
from nzpyida.analytics.utils import get_auto_delete_context
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling


class KMeans(PredictiveModeling):
    """
    KMeans clustering.
    """

    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the clusterer class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.fit_proc = 'KMEANS'
        self.predict_proc = 'PREDICT_KMEANS'
        self.score_proc = 'MSE'
        self.target_column_in_output = 'CLUSTER_ID'
        self.id_column_in_output = 'ID'
        self.has_print_proc = True

    def fit(self, in_df: IdaDataFrame, id_column: str=None,
        in_columns: List[str]=None, col_def_type: str=None, col_def_role: str=None,
        col_properties_table: str=None, out_table: str=None, distance: str='norm_euclidean',
        k: int=3, max_iter: int=5, rand_seed: int=12345, id_based: bool=False,
        statistics: str=None, transform: str='L') -> IdaDataFrame:
        """
        Creates and trains a model for clustering based on provided data and store
        it in a database.

        The training algorithm operates by performing several iterations of the same
        basic process. Each training instance is assigned to the closest cluster
        with respect to the specified distance function, applied to the instance
        and cluster center. All cluster centers are then re-calculated as the mean
        attribute value vectors of the instances assigned to particular clusters.
        The cluster centers are initialized by randomly picking k training instances,
        where k is the desired number of clusters. The iterative process should
        terminate when there are either no or sufficiently few changes in cluster
        assignments. In practice, however, it is sufficient to specify the number of
        iterations, typically a number between 3 and 36.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        in_columns : List[str], optional
            the list of input table columns with special properties.
            Each column is followed by one or several of the following properties:
                its type: ':nom' (for nominal), ':cont' (for continuous).
                Per default, all numerical types are con-tinuous, other types are nominal.
                its role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same
            as ':colweight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input table column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')

        out_table : str, optional
            the output table where clusters are assigned to each input table record

        distance : str, optional
            the distance function. Allowed values are: euclidean, norm_euclidean, manhattan,
            canberra, maximum, mahalanobis.

        k : int, optional
            number of centers

        max_iter : int, optional
            the maximum number of iterations to perform

        rand_seed : int, optional
            the random generator seed

        id_based : bool, optional
            the specification that random generator seed is based on id column value

        statistics : str, optional
            flags indicating which statistics to collect.
            Allowed values are: none, columns, values:n, all.
            If statistics=none, no statistics are collected.
            If statistics=columns, statistics on the input table columns like mean value are
            collec-ted.
            If statistics=values:n with n a positive number, statistics about the columns and
            the column values are collected. Up to <n> column value statistics are collected:
            If a nominal column contains more than <n> values, only the <n> most frequent
            column statistics are kept.
            If a numeric column contains more than <n> values, the values will be discretized and
            the statistics will be collected on the discretized values.
            Indicating statistics=all is equal to statistics=values:100.

        transform : str, optional
            flag indicating if the input table columns have to be transformed.
            Allowed values are: L (for leave as is), N (for normalization)
            or S (for standardization).
            If it is not specified, no transformation will be performed.
        
        Returns
        -------
        IdaDataFrame
            output table with following columns: id, cluster_id, distance. The id column matches 
            the <id_column>  of the input table. Each input table record is associated with a cluster, 
            where the distance from the record to the cluster center is the smallest. The cluster ID 
            and the distance to the cluster center are given in the columns cluster_id and distance

        """
        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        if not id_column:
            if in_df.indexer:
                id_column = in_df.indexer
            else:
                raise TypeError('Missing id column - either use id_column attribute or set '
                    'indexer column in the input data frame')

        auto_delete_context = None
        if not out_table:
            auto_delete_context = get_auto_delete_context('out_table')
            out_table = make_temp_table_name()

        params = {
            'id': id_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
            'distance': distance,
            'k': k,
            'maxiter': max_iter,
            'randseed': rand_seed,
            'idbased': id_based,
            'statistics': statistics,
            'transform': transform,
            'outtable': out_table
        }

        self._fit(in_df=in_df, params=params)

        if auto_delete_context:
            auto_delete_context.add_table_to_delete(out_table)

        return IdaDataFrame(self.idadb, out_table)

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table where the assigned clusters will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """

        params = {
            'id': id_column
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)

    def score(self, in_df: IdaDataFrame, target_column: str,
        id_column: str=None) -> float:
        """
        Scores the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        Returns
        -------
        float
            the model score
        """

        params = {
            'id': id_column
        }

        return self._score(in_df=in_df, predict_params=params, target_column=target_column)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
import numpy as np
from nzpyida.base import IdaDataBase
from nzpyida.analytics.auto_delete_context import AutoDeleteContext
from nzpyida.analytics.tests.conftest import df_train, TAB_NAME_TRAIN, \
    TAB_NAME_TEST
from nzpyida.analytics.transform.preparation import std_norm, impute_data, \
    random_sample, train_test_split

TAB_NAME = "TAB_NAME0"

@pytest.fixture(scope='function')
def clean_up(idadb: IdaDataBase):
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)
    if idadb.exists_table(TAB_NAME_TEST):
        idadb.drop_table(TAB_NAME_TEST)
    yield
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)
    if idadb.exists_table(TAB_NAME_TEST):
        idadb.drop_table(TAB_NAME_TEST)

@pytest.fixture(scope='module')
def idf(idadb: IdaDataBase):
    idf = idadb.as_idadataframe(df_train, tablename=TAB_NAME, clear_existing=True, indexer='ID')
    yield idf
    if idadb.exists_table(TAB_NAME):
        idadb.drop_table(TAB_NAME)

def test_std_norm(idadb: IdaDataBase, clean_up, idf):

    out_df = std_norm(idf, in_column=["A:S"], by_column=['B'], out_table=TAB_NAME_TEST)
    assert out_df
    assert idadb.exists_table_or_view(TAB_NAME_TEST)

    assert all(out_df.columns == ['B', 'ID', 'STD_A'])

    assert len(out_df) == len(idf)


def test_random_sample(idadb: IdaDataBase, idf, clean_up):
    with AutoDeleteContext(idadb):
        sample_df = random_sample(idf, size=300, by_column=["B"], rand_seed=123)
        assert sample_df

        assert all(sample_df.columns == [ 'ID', 'A', 'B'])

        assert len(sample_df) == 300

        sample_df2 = random_sample(idf, size=300, by_column=["B"], rand_seed=123)
        sample_df3 = random_sample(idf, size=300, by_column=["B"])
        assert all((sample_df.head(10) == sample_df2.head(10))['ID'])
        assert not all((sample_df.head(10) == sample_df3.head(10))['ID'])


def test_impute_data(idadb: IdaDataBase, clean_up):
    df_train['new'] = np.nan
    idf = idadb.as_idadataframe(df_train, tablename=TAB_NAME_TRAIN, clear_existing=True, indexer='ID')

    assert idf
    assert idadb.exists_table_or_view(TAB_NAME_TRAIN)


def test_train_test_split(idadb: IdaDataBase, idf, clean_up):
    train_df, test_df = train_test_split(idf, TAB_NAME_TRAIN, TAB_NAME_TEST, fraction=0.8,
                                         id_column="ID")
    assert test_df
    assert test_df
    assert round(len(train_df)/len(idf), 1) == 0.8
    assert round(len(test_df)/len(idf), 1) == 0.2

    assert not any(el in train_df['ID'].as_dataframe().values for el in test_df["ID"].as_dataframe().values)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
TwoStep clustering is a data mining algorithm for large data sets. It is faster than 
traditional methods because it typically scans a data set only once before it saves 
the data to a clustering feature (CF) tree. TwoStep clustering can make clustering 
decisions without repeated data scans, whereas other clustering methods scan all 
data points, which requires multiple iterations. Non- uniform points are not gathered, 
so each iteration requires a reinspection of each data point, regardless of the 
significance of the data point. Because TwoStep clustering treats dense areas 
as a single unit and ignores pattern outliers, it provides high-quality clustering 
results without exceeding memory constraints.

The TwoStep algorithm has the following advantages:
- It automatically determines the optimal number of clusters. You do not have to 
manually create a different clustering model for each number of clusters.
- It detects input columns that are not useful for the clustering process. 
These columns are automatically set to supplementary. Statistics are gathered 
for these columns but they do not influence the clustering algorithm.
- The configuration of the CF tree can be granular, so that you can balance between 
memory usage and model quality, according to the environment and needs.
"""

from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, make_temp_table_name
from nzpyida.analytics.utils import get_auto_delete_context
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling

class TwoStepClustering(PredictiveModeling):
    """
    Divisive Clustering
    """
    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the clusterer class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.fit_proc = 'TWOSTEP'
        self.predict_proc = 'PREDICT_TWOSTEP'
        self.score_proc = 'MSE'
        self.target_column_in_output = 'CLUSTER_ID'
        self.id_column_in_output = 'ID'
        self.has_print_proc = True

    def fit(self, in_df: IdaDataFrame, id_column: str=None, target_column: str=None,
        in_columns: List[str]=None, col_def_type: str=None, col_def_role: str=None,
        col_properties_table: str=None, out_table: str=None, k: int=0, max_k: int=20, 
        bins: int=10, statistics: str=None, rand_seed: int=12345, distance: str='loglikelihood', 
        distance_threshold: float=None, distance_threshold_factor: float=2.0, 
        epsilon: float=0.0, node_capacity: int=6, leaf_capacity: int=8, 
        max_leaves: int=1000, outlier_fraction: float=0.0) -> IdaDataFrame:
        """
        Builds a TwoStep Clustering model that first distributes the input data into 
        a hierarchical tree structure according to the distance between the data records, 
        then reduces the tree into k clusters. A second pass over the data associates 
        the input data records to the next cluster.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        id_column : str, optional
            the input table column identifying a unique instance id
        
        target_column : str, optional
            the input table column representing a class or a value to predict, 
            this column is ignored by the TwoStep Clustering algorithm

        in_columns : List[str], optional
            the list of input table columns with special properties.
            Each column is followed by one or several of the following properties:
                its type: ':nom' (for nominal), ':cont' (for continuous).
                Per default, all numerical types are con-tinuous, other types are nominal.
                its role: ':id', ':target', ':input', ':ignore'.
            (Remark: ':objweight' is unsupported, i.e. ':objweight' same as ':ignore').
            (Remark: ':colweight(<wgt>)' is unsupported, i.e. ':colweight(<wgt>)' same
            as ':colweight(1)' same as ':input').
            If the parameter is undefined, all columns of the input table have default properties.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns.
            Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input' columns.

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            The format of this table is the output format of stored procedure
            nza..COLUMN_PROPERTIES().
            If the parameter is undefined, the input table column properties will be
            detected automatically.
            (Remark: colPropertiesTable with "COLROLE" column with value 'objweight'
            is unsupported, i.e. same as 'ignore')
            (Remark: colPropertiesTable with "COLWEIGHT" column with value '<wgt>' is unsupported,
            i.e. same as '1')

        out_table : str, optional
            the output table where clusters are assigned to each input table record
        
        k : int, optional
            the number of clusters. If k is 0 or less, the procedure determines 
            the optimal number of clusters

        max_k : int, optional
            the maximum number of clusters that can be determined automatically. 
            If k is bigger than 0, this parameter is ignored
        
        bins : int, optional
            the average number of bins for numerical statistics with more than <n> values

        statistics : str, optional
            flags indicating which statistics to collect. Allowed values are: none, 
            columns, values:n, all.
            Regardless of the value of the parameter statistics, all statistics are 
            gathered since they are needed to call PREDICT_TWOSTEP on this model. 
            If statistics=none or statistics=columns, the importance of the attributes 
            is not calculated. If statistics=none, statistics=columns or statistics=all, 
            up to 100 discrete values are gathered.
            If statistics=values:n with n a positive number, up to <n> column value 
            statistics are collected:
                - If a nominal column contains more than <n> values, only the <n> most 
                frequent column stat-istics are kept.
                - If a numeric column contains more than <n> values, the values will 
                be discretized and the stat-istics will be collected on the discretized values.
            Indicating statistics=all is equal to statistics=values:100.
            
        rand_seed : int, optional
            the random generator seed

        distance : str, optional
            the distance function. Allowed values are: euclidean, norm_euclidean, loglikelihood

        distance_threshold : float, optional
            the threshold under which 2 data records can be merged into one cluster during 
            the first pass. If not set, the distance threshold is calculated automatically
        
        distance_threshold_factor : float, optional
            the factor used to calculate the distance threshold automatically. 
            The distance threshold is then the median distance value minus 
            distance_threshold_factor times the interquartile distance 
            (or the minimum distance if this value is below it). If distance_threshold is set, 
            this parameter is ignored
        
        epsilon : float, optional
            the value to be used as global variance of all continuous fields for the loglikelihood 
            distance. If the value is 0.0 or less, the global variance is calculated for each 
            continuous field. If distance is not loglikelihood, this parameter is ignored
        
        node_capacity : int, optional
            the branching factor of the internal tree used in pass 1. Each node can have up to 
            node_capacity subnodes
        
        leaf_capacity : int, optional
            the number of clusters per leaf node in the internal tree used in pass 1

        max_leaves : int, optional
            the maximum number of leaf nodes in the internal tree used in pass 1. 
            When the tree contains maxleaves leaf nodes, the following data records are 
            aggregated into the existing clusters
        
        outlier_fraction : float, optional
            the fraction of the records to be considered as outlier in the internal
            tree used in pass 1. Clusters containing less than outlierfraction times 
            the mean number of data records per cluster are removed
        
        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers, cluster_id and distance to cluster center
        """

        auto_delete_context = None
        if not out_table:
            auto_delete_context = get_auto_delete_context('out_table')
            out_table = make_temp_table_name()

        params = {
            'id': id_column,
            'target': target_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
            'k': k,
            'maxk': max_k,
            'bins': bins,
            'statistics': statistics,
            'randseed': rand_seed,
            'distance': distance,
            'distancethreshold': distance_threshold,
            'distancethresholdfactor': distance_threshold_factor,
            'epsilon': epsilon,
            'nodecapacity': node_capacity,
            'leafcapacity': leaf_capacity,
            'maxleaves': max_leaves,
            'outlierfraction': outlier_fraction,
            'outtable': out_table
        }

        self._fit(in_df=in_df, params=params)

        if auto_delete_context:
            auto_delete_context.add_table_to_delete(out_table)

        return IdaDataFrame(self.idadb, out_table)

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table where the assigned clusters will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """

        params = {
            'id': id_column
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)

    def score(self, in_df: IdaDataFrame, target_column: str, id_column: str=None) -> float:
        """
        Scores the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        Returns
        -------
        float
            the model score
        """

        params = {
            'id': id_column
        }

        return self._score(in_df=in_df, predict_params=params, target_column=target_column)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida import IdaDataBase, IdaDataFrame
from nzpyida.analytics.auto_delete_context import AutoDeleteContext
import pytest
from nzpy.core import ProgrammingError
import importlib

new_table_command = """
drop table test_table if exists;
CREATE TABLE test_table (col1 int, col2 int, col3 int);
insert into test_table (col1, col2, col3) values (11, 22, 33);
"""
new_table_command_2="""
DROP TABLE TEST_TABLE_2 IF EXISTS;
CREATE TABLE test_table_2 (col1 int, col2 int, col3 int);
insert into test_table_2 (col1, col2, col3) values (12, 23, 34);
"""


@pytest.fixture
def create_test_table(idadb: IdaDataBase):
    idadb.ida_query(new_table_command)
    yield
    idadb.ida_query("DROP TABLE TEST_TABLE IF EXISTS")

@pytest.fixture
def create_test_table_2(idadb: IdaDataBase):
    idadb.ida_query(new_table_command)
    idadb.ida_query(new_table_command_2)
    yield
    idadb.ida_query("DROP TABLE TEST_TABLE IF EXISTS")
    idadb.ida_query("DROP TABLE TEST_TABLE_2 IF EXISTS")


class TestCurrentAutoDeleteContext:
    def test_context_present(self, idadb: IdaDataBase):
        with AutoDeleteContext(idadb) as adc:
            assert adc.current()

    def test_context_not_present(self, idadb):
        with AutoDeleteContext(idadb) as adc:
            pass
        assert not adc.current()

    def test_nested_contexts(self, idadb):
        with AutoDeleteContext(idadb) as adc1:
            with AutoDeleteContext(idadb) as adc2:
                assert adc1 != adc2
                assert adc2 == AutoDeleteContext.current()
            assert adc1 == AutoDeleteContext.current()


class TestAddTableToDelete:
    def test_table_not_deleted(self, idadb, create_test_table):
        with AutoDeleteContext(idadb) as adc:
            df = IdaDataFrame(idadb, 'TEST_TABLE')
            adc.add_table_to_delete('TEST_TABLE')
            assert df

    def test_table_deleted(self, idadb, create_test_table):
        with pytest.raises(ProgrammingError, match=r".*relation does not exist*"):
            with AutoDeleteContext(idadb) as adc:
                df = IdaDataFrame(idadb, 'TEST_TABLE')
                adc.add_table_to_delete('TEST_TABLE')
            assert df
    
    def test_double_tables_deleted(self, idadb, create_test_table_2):
        with AutoDeleteContext(idadb) as adc:
            df = IdaDataFrame(idadb, 'TEST_TABLE')
            df2 = IdaDataFrame(idadb, 'TEST_TABLE_2')
            adc.add_table_to_delete('TEST_TABLE')
            adc.add_table_to_delete('TEST_TABLE_2')
        assert not any([df.exists(), df2.exists()])

    def test_double_table_one_not_deleted(self, idadb, create_test_table_2):
        with AutoDeleteContext(idadb) as adc:
            df = IdaDataFrame(idadb, 'TEST_TABLE')
            df2 = IdaDataFrame(idadb, 'TEST_TABLE_2')
            adc.add_table_to_delete('TEST_TABLE')
        assert not df.exists() and df2.exists()

    def test_double_tables_deleted(self, idadb, create_test_table_2):
        with AutoDeleteContext(idadb) as adc:
            df = IdaDataFrame(idadb, 'TEST_TABLE')
            df2 = IdaDataFrame(idadb, 'TEST_TABLE_2')
            adc.add_table_to_delete('TEST_TABLE')
            adc.add_table_to_delete('TEST_TABLE_2')
        assert not any([df.exists(), df2.exists()])

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.analytics.predictive.glm import BernoulliRegressor, NegativeBinomialRegressor, BinomialRegressor, \
    GaussianRegressor, GammaRegressor, PoissonRegressor, WaldRegressor
from nzpyida.analytics.tests.conftest import MOD_NAME, OUT_TABLE_PRED, OUT_TABLE_CM, TAB_NAME_TEST, \
    TAB_NAME_TRAIN, df_test_reg, df_train_reg
import pandas as pd
import pytest

@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture(scope='function')
def clear_up(idadb: IdaDataBase, mm: ModelManager):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)

def model_family(idadb: IdaDataBase, model_name: str):
    models = {
        'bernoulli': {"model": BernoulliRegressor(idadb, MOD_NAME),
                     "params": {'max_iter': 15}},
        'gaussian': {"model": GaussianRegressor(idadb, MOD_NAME),
                     "params": {}},
        'poisson': {"model": PoissonRegressor(idadb, MOD_NAME),
                     "params": {}},
        'binomial': {"model": BinomialRegressor(idadb, MOD_NAME),
                     "params": {'trials': "T"}},
        'negativebinomial': {"model": NegativeBinomialRegressor(idadb, MOD_NAME),
                     "params": {}},
        'wald': {"model": WaldRegressor(idadb, MOD_NAME),
                     "params": {}},
        'gamma': {"model": GammaRegressor(idadb, MOD_NAME),
                     "params": {}},
    }
    return models[model_name]

@pytest.mark.parametrize("model_name", ['bernoulli', 'gaussian', 'poisson', 
                                        'binomial', 'negativebinomial', 'wald', 'gamma'
                                        ])
def test_glm(idadb: IdaDataBase, mm: ModelManager, clear_up, model_name):
    idf_train = idadb.as_idadataframe(df_train_reg, tablename=TAB_NAME_TRAIN, clear_existing=True)
    idf_test = idadb.as_idadataframe(df_test_reg, tablename=TAB_NAME_TEST, clear_existing=True)
    assert idf_train
    assert idf_test
    
    model_info = model_family(idadb, model_name)
    
    model = model_info["model"]
    assert model
    assert not mm.model_exists(MOD_NAME)

    params = {
        'in_df': idf_train,
        'id_column': "ID",
        'target_column': "B",
        'in_columns': None,  
        'intercept': True, 
        'interaction': '', 
        'family_param': -1, 
        'link': 'logit', 
        'link_param': 1, 
        'max_iter': 20, 
        'epsilon': 1e-3, 
        'tolerance': 1e-7, 
        'method': 'irls', 
        'trials': '', 
        'debug': False, 
        'col_def_type': None, 
        'col_def_role': None, 
        'col_properties_table': None
    }
    for k, v in model_info["params"].items():
        params[k] = v
    model.fit(**params)
    assert mm.model_exists(MOD_NAME)

    pred = model.predict(idf_test, id_column="ID", out_table=OUT_TABLE_PRED)
    assert pred
    assert all(pred.columns == ['ID', 'PRED'])
    # assert any(round(x) == y for x, y in zip(list(pred.as_dataframe()['CLASS'].values),  [1, 4, 2223, -999, 11112]))

    score = model.score(idf_test, id_column="ID", target_column="B")

    assert score

    mse, mae, rse, rae = model.score_all(idf_test, id_column='ID', target_column='B')
    assert all([mse, mae, rse, rae])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
The nearest neighbor family of classification and regression algorithms is
frequently referred to as memory-based or instance-based learning, and
sometimes also as lazy learning. These terms correspond to the main concept
of this approach, which is to replace model creation by memorizing the
training data set and using it appropriately to make predictions.
"""
from typing import List
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.predictive.classification import Classification


class KNeighborsClassifier(Classification):
    """
    K-neighbors based classifier.
    """

    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the classifier class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """
        super().__init__(idadb, model_name)
        self.fit_proc = 'KNN'
        self.predict_proc = 'PREDICT_KNN'

    def fit(self, in_df: IdaDataFrame, target_column: str, id_column: str=None,
        in_columns: List[str]=None, col_def_type: str=None, col_def_role: str=None,
        col_properties_table: str=None):
        """
        Builds a K-Nearest Neighbors Classification or Regression model.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        in_columns : List[str], optional
            the list of input table columns with special properties.
            Each column is followed by one or several of the following properties:
            its type: ':nom' (for nominal), ':cont' (for continuous).
                Per default, all numerical types are continuous, other types are nominal.
            its role: ':id', ':target', ':input', ':ignore'.

        col_def_type : str, optional
            default type of the input table columns. Allowed values are 'nom' and 'cont'.
            If the parameter is undefined, all numeric columns are considered continuous,
            other columns nominal.

        col_def_role : str, optional
            default role of the input table columns. Allowed values are 'input' and 'ignore'.
            If the parameter is undefined, all columns are considered 'input'

        col_properties_table : str, optional
            the input table where column properties for the input table columns are stored.
            If the parameter is undefined, the input table column properties will be detected
            automatically.
        """

        params = {
            'id': id_column,
            'target': target_column,
            'incolumn': in_columns,
            'coldeftype': col_def_type,
            'coldefrole': col_def_role,
            'colpropertiestable': col_properties_table,
        }

        self._fit(in_df=in_df, params=params)

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None, distance: str='euclidean', k: int=3, stand: bool=True,
        fast: bool=True, weights: str=None) -> IdaDataFrame:
        """
        Applies a K-Nearest Neighbors model to generate classification or regression
        predictions for a data frame.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame

        out_table : str, optional
            the output table where the predictions will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model

        distance : str, optional
            the distance function. Allowed values are: euclidean, manhatthan, canberra, maximum

        k : int, optional
            number of nearest neighbors to consider

        stand : bool, optional
            flag indicating whether the measurements in the input table are standardized before
            calculating the distance

        fast : bool, optional
            flag indicating that the algorithm used coresets based method

        weights : str, optional
            the input table containing optional class weights for the input table <target> column.
            The <weights> table is used only when the <target> column is not numeric.
            If the parameter is undefined, we assume that the weights are uniformly equal to 1.
            The <weights> table contains following columns:
                weight: a numeric column containing the class weight,
                class: a column to be joined with the <target> column of <intable>, defining class
                weights.
            For classes not occurring in this table, weights of 1 are assumed.

        Returns
        -------
        IdaDataFrame
            a data frame with id and predicted class

        """

        params = {
            'id': id_column,
            'distance': distance,
            'k': k,
            'stand': stand,
            'fast': fast,
            'weights': weights
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)

    def score(self, in_df: IdaDataFrame, target_column: str, id_column: str=None,
        distance: str='euclidean', k: int=3, stand: bool=True, fast: bool=True,
        weights: str=None) -> float:
        """
        Scores the model and returns classification error ratio.

        Parameters
        ----------

        in_df : IdaDataFrame
            the input data frame used to test the model

        target_column : str
            the input table column representing the class in the input data frame

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        distance : str, optional
            the distance function. Allowed values are: euclidean, manhatthan, canberra, maximum

        k : int, optional
            number of nearest neighbors to consider

        stand : bool, optional
            flag indicating whether the measurements in the input table are standardized before
            calculating the distance

        fast : bool, optional
            flag indicating that the algorithm used coresets based method

        weights : str, optional
            the input table containing optional class weights for the input table <target> column.
            The <weights> table is used only when the <target> column is not numeric.
            If the parameter is undefined, we assume that the weights are uniformly equal to 1.
            The <weights> table contains following columns:
                weight: a numeric column containing the class weight,
                class: a column to be joined with the <target> column of <intable>, defining class
                weights.
            For classes not occurring in this table, weights of 1 are assumed.

        Returns
        -------
        float
            model classification error ratio
        """

        params = {
            'id': id_column,
            'target': target_column,
            'distance': distance,
            'k': k,
            'stand': stand,
            'fast': fast,
            'weights': weights
        }

        return self._score(in_df=in_df, predict_params=params, target_column=target_column)

    def conf_matrix(self, in_df: IdaDataFrame, target_column: str, id_column: str=None, out_matrix_table: str = None,
                    distance: str='euclidean', k: int=3, stand: bool=True, fast: bool=True, weights: str=None):
        """
        Makes a predition for a test data set given by the user and returns a confusion matrix,
        together with other stats (ACC and WACC).

        Parameters
        -----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id

        out_matrix_table : str, optional
            the output table where the confidence matrix will be stored

        Returns
        --------
        IdaDataFrame
            the confidence matrix data frame

        float
            classification accuracy (ACC)

        float
            weighted classification accuracy (WACC)
        """
        params = {
            'id': id_column,
            'target': target_column,
            'distance': distance,
            'k': k,
            'stand': stand,
            'fast': fast,
            'weights': weights
        }

        return self._conf_matrix(in_df=in_df, params=params, out_matrix_table=out_matrix_table)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.analytics.predictive.regression_trees import DecisionTreeRegressor
from nzpyida.analytics.tests.conftest import MOD_NAME, OUT_TABLE_PRED, OUT_TABLE_CM, TAB_NAME_TEST, \
    TAB_NAME_TRAIN, df_test_reg, df_train_reg
import pandas as pd
import pytest

@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture(scope='module')
def clear_up(idadb: IdaDataBase, mm: ModelManager):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CM):
        idadb.drop_table(OUT_TABLE_CM)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CM):
        idadb.drop_table(OUT_TABLE_CM)

def test_regression_trees(idadb: IdaDataBase, mm: ModelManager, clear_up):
    idf_train = idadb.as_idadataframe(df_train_reg, tablename=TAB_NAME_TRAIN, clear_existing=True)
    idf_test = idadb.as_idadataframe(df_test_reg, tablename=TAB_NAME_TEST, clear_existing=True)
    assert idf_train
    assert idf_test

    model = DecisionTreeRegressor(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME)
    
    model.fit(idf_train, id_column="ID", target_column="B", min_improve=0.001, min_split=200)
    assert mm.model_exists(MOD_NAME)

    pred = model.predict(idf_test, id_column="ID", out_table=OUT_TABLE_PRED)
    assert pred
    assert all(pred.columns == ['ID', 'CLASS'])
    # assert any(round(x) == y for x, y in zip(list(pred.as_dataframe()['CLASS'].values),  [1, 4, 2223, -999, 11112]))

    score = model.score(idf_test, id_column="ID", target_column="B")

    assert score

    mse, mae, rse, rae = model.score_all(idf_test, id_column='ID', target_column='B')
    assert all([mse, mae, rse, rae])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
This module contains a class that is the base for all classification algorithms.
"""
from typing import Tuple
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, materialize_df, make_temp_table_name, \
call_proc_df_in_out
from nzpyida.analytics.utils import get_auto_delete_context
from nzpyida.analytics.predictive.predictive_modeling import PredictiveModeling
from nzpyida.analytics.model_manager import ModelManager


class Classification(PredictiveModeling):
    """
    Base class for classification algorithms.
    """

    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the classifier class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        super().__init__(idadb, model_name)
        self.target_column_in_output = 'CLASS'
        self.id_column_in_output = 'ID'
        self.score_proc = 'CERROR'
        self.score_inv = True
        self.type = None

    def predict(self, in_df: IdaDataFrame, out_table: str=None,
        id_column: str=None) -> IdaDataFrame:
        """
        Makes predictions based on this model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for predictions

        out_table : str, optional
            the output table where the predictions will be stored

        id_column : str, optional
            the input table column identifying a unique instance id
            Default: id column used to build the model
        """

        params = {
            'id': id_column
        }

        return self._predict(in_df=in_df, params=params, out_table=out_table)

    def score(self, in_df: IdaDataFrame, target_column: str, id_column: str=None) -> float:
        """
        Scores the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        Returns
        -------
        float
            the model score
        """

        params = {
            'id': id_column
        }

        return self._score(in_df=in_df, predict_params=params, target_column=target_column)

    def conf_matrix(self, in_df: IdaDataFrame, target_column: str, id_column: str=None,
        out_matrix_table: str=None) -> Tuple[IdaDataFrame, float, float]:
        """
        Makes a predition for a test data set given by the user and returns a confusion matrix,
        together with other stats (ACC and WACC).

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        target_column : str
            the input table column representing the class

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        out_matrix_table : str, optional
            the output table where the confidence matrix will be stored

        Returns
        -------
        IdaDataFrame
            the confidence matrix data frame

        float
            classification accuracy (ACC)

        float
            weighted classification accuracy (WACC)
        """

        params = {
            'id': id_column,
            'target': target_column
        }
        return self._conf_matrix(in_df, out_matrix_table, params)
        
    def _conf_matrix(self, in_df: IdaDataFrame, out_matrix_table: str=None,
                     params: dict={}) -> Tuple[IdaDataFrame, float, float]:

        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        if not params.get('id'):
            if in_df.indexer:
                params['id'] = in_df.indexer
            else:
                raise TypeError('Missing id column - either use id_column attribute or set '
                    'indexer column in the input data frame')

        out_table = make_temp_table_name()

        pred_view_needs_delete, true_view_needs_delete = False, False
        try:
            pred_df = self._predict(in_df=in_df, out_table=out_table, params=params)
            pred_view, pred_view_needs_delete = materialize_df(pred_df)
            true_view, true_view_needs_delete = materialize_df(in_df)

            auto_delete_context = None
            if not out_matrix_table:
                auto_delete_context = get_auto_delete_context('out_matrix_table')
                out_matrix_table = make_temp_table_name()

            params_s = map_to_props({
                'resulttable': pred_view,
                'intable': true_view,
                'resultid': 'ID',
                'id': params['id'],
                'resulttarget': 'CLASS',
                'target': params['target'],
                'matrixTable': out_matrix_table
            })
            self.idadb.ida_query(f'call NZA..CONFUSION_MATRIX(\'{params_s}\')')

            if auto_delete_context:
                auto_delete_context.add_table_to_delete(out_matrix_table)

            out_df = IdaDataFrame(self.idadb, out_matrix_table)

            params = map_to_props({
                'matrixTable': out_matrix_table
            })

            res_acc = self.idadb.ida_query(f'call NZA..CMATRIX_ACC(\'{params}\')')
            res_wacc = self.idadb.ida_query(f'call NZA..CMATRIX_WACC(\'{params}\')')

            return out_df, res_acc[0], res_wacc[0]

        finally:
            self.idadb.drop_table(out_table)
            if pred_view_needs_delete:
                self.idadb.drop_view(pred_view)
            if true_view_needs_delete:
                self.idadb.drop_view(true_view)
    
    def cross_validation(self, in_df: IdaDataFrame, id_column: str=None, 
                         target_column: str=None, out_table: str=None, folds: int=10, 
                         rand_seed: float=None) -> Tuple[IdaDataFrame, float]:
        """
        Performs a cross validation on <in_df> data for given model. Numer of batches 
        and size of train/test split isdetermined by parameter <folds>

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame for scoring

        id_column : str, optional
            the input table column identifying a unique instance id - if skipped, 
            the input data frame indexer must be set and will be used as an instance id

        target_column : str
            the input table column representing the class

        out_table : str, optional
            the output table where the predicted values will be stored

        Returns
        -------
        IdaDataFrame
            the data frame with predicted values for all <in_df>

        float
            classification accuracy (ACC) for all batches 
        """
        params = {
            'modelType': self.fit_proc,
            'model': self.model_name,
            'intable': in_df,
            'id': id_column,
            'target': target_column,
            'outtable': out_table,
            'folds': folds,
        }
        if isinstance(rand_seed, int):
            params['seed'] = rand_seed
        
        ModelManager(self.idadb).drop_model(self.model_name)
        
        ret_df, ret_acc = call_proc_df_in_out(proc="CROSS_VALIDATION", in_df=in_df, params=params,
                                   out_table=out_table)
        
        return ret_df, ret_acc[0]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
This module contains a class that is the base for all predictive algorithms.
"""
from nzpyida.frame import IdaDataFrame
from nzpyida.base import IdaDataBase
from nzpyida.analytics.utils import map_to_props, materialize_df, make_temp_table_name
from nzpyida.analytics.utils import call_proc_df_in_out
from nzpyida.analytics.model_manager import ModelManager


class PredictiveModeling:
    """
    Generic class for predictive modeling algorithms.
    """

    def __init__(self, idadb: IdaDataBase, model_name: str):
        """
        Creates the predictive modeling class.

        Parameters
        ----------

        idada : IdaDataBase
            database connector

        model_name : str
            model name - if it exists in the database, it will be used, otherwise
            it must be trained using fit() function before prediction or scoring is called.
        """

        self.idadb = idadb
        self.model_name = model_name
        self.fit_proc = ''
        self.predict_proc = ''
        self.score_proc = ''
        self.score_inv = False
        self.target_column_in_output = None
        self.id_column_in_output = None
        self.has_print_proc = False

    def _fit(self, in_df: IdaDataFrame, params:dict, needs_id=True):
        """
        Trains the model.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        params : dict
            the dictionary of attributes used to build the model
        """
        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        if not params.get('id', None) and needs_id:
            if in_df.indexer:
                params['id'] = in_df.indexer
            else:
                raise TypeError('Missing id column - either use id_column attribute or set '
                    'indexer column in the input data frame')

        ModelManager(self.idadb).drop_model(self.model_name)

        temp_view_name, need_delete = materialize_df(in_df)

        params['model'] = self.model_name
        params['intable'] = temp_view_name
        params_s = map_to_props(params)

        try:
            self.idadb.ida_query(f'call NZA..{self.fit_proc}(\'{params_s}\')')
        finally:
            if need_delete:
                self.idadb.drop_view(temp_view_name)

    def _predict(self, in_df: IdaDataFrame, params:dict, out_table: str=None) -> IdaDataFrame:
        """
        Makes predictions based on the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        params : dict
            the dictionary of attributes used for making predictions

        out_table : str, optional
            the output table where the predictions will be stored

        Returns
        -------
        IdaDataFrame
            the data frame containing row identifiers and predicted target values
        """
        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        params['model'] = self.model_name
        return call_proc_df_in_out(proc=self.predict_proc, in_df=in_df, params=params,
            out_table=out_table)[0]

    def _score(self, in_df: IdaDataFrame, predict_params:dict, target_column: str) -> float:
        """
        Scores the model. The model must exist.

        Parameters
        ----------
        in_df : IdaDataFrame
            the input data frame

        predict_params : dict
            the dictionary of attributes used for making predictions

        target_column : str
            the input table column representing the class

        Returns
        -------
        float
            the model score
        """
        if not isinstance(in_df, IdaDataFrame):
            raise TypeError("Argument in_df should be an IdaDataFrame")

        if not predict_params.get('id', None):
            if in_df.indexer:
                predict_params['id'] = in_df.indexer
            else:
                raise TypeError('Missing id column - either use id_column attribute or set '
                    'indexer column in the input data frame')

        out_table = make_temp_table_name()

        pred_view_needs_delete, true_view_needs_delete = False, False
        try:

            pred_df = self._predict(in_df=in_df, params=predict_params, out_table=out_table)

            pred_view, pred_view_needs_delete = materialize_df(pred_df)
            true_view, true_view_needs_delete = materialize_df(in_df)

            id_column = predict_params.get('id')

            params = map_to_props({
                'pred_table': pred_view,
                'true_table': true_view,
                'pred_id': id_column if self.id_column_in_output is None
                    else self.id_column_in_output,
                'true_id': id_column,
                'pred_column': target_column if self.target_column_in_output is None
                    else self.target_column_in_output,
                'true_column': target_column
            })

            res = self.idadb.ida_query(f'call NZA..{self.score_proc}(\'{params}\')')
            return 1-res[0] if self.score_inv else res[0]
        finally:
            self.idadb.drop_table(out_table)
            if pred_view_needs_delete:
                self.idadb.drop_view(pred_view)
            if true_view_needs_delete:
                self.idadb.drop_view(true_view)

    def describe(self) -> str:
        """
        Returns model description.

        Returns
        -------
        str
            model description
        """
        if self.has_print_proc:
            params = map_to_props({'model': self.model_name})
            return self.idadb.ida_query(f'call NZA..PRINT_MODEL(\'{params}\')')[0]
        return ''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from nzpyida.analytics.predictive.association_rules import ARule
from nzpyida.base import IdaDataBase
from nzpyida.frame import IdaDataFrame
from nzpyida.analytics.model_manager import ModelManager
import pytest
from nzpyida.analytics.tests.conftest import MOD_NAME, TAB_NAME_TRAIN, \
    TAB_NAME_TEST, OUT_TABLE_PRED
import pandas as pd
from random import choice

@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture
def clean_up(idadb, mm):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)



@pytest.fixture
def idf_train(idadb: IdaDataBase):
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)
    purchases = [[1,2,3,4,5],
                 [2,4,6,7,10],
                 [3,5,8,9,10],
                 [4,6,7,8,10],
                 [1,2,3,7,10],
                 [2,3,4,6,9],
                 [5,6,7,8,10],
                 [1,3,4,6,9],
                 [1,2,3,8,9],
                 [3,5,7,9,10]]
    df = pd.DataFrame.from_dict({
        "ID": [ i//5 for i in range(50)],
        "PRODUCT": [item for sublist in purchases for item in sublist]

    })
    yield idadb.as_idadataframe(df, TAB_NAME_TRAIN)
    if idadb.exists_table(TAB_NAME_TRAIN):
        idadb.drop_table(TAB_NAME_TRAIN)

@pytest.fixture
def idf_test(idadb: IdaDataBase):
    if idadb.exists_table(TAB_NAME_TEST):
        idadb.drop_table(TAB_NAME_TEST)
    df = pd.DataFrame.from_dict({
        "TID": range(50),
        "ITEM": [choice(range(10)) for _ in range(50)]
    })
    yield idadb.as_idadataframe(df, TAB_NAME_TEST)
    if idadb.exists_table(TAB_NAME_TEST):
        idadb.drop_table(TAB_NAME_TEST)

def test_arule(idadb: IdaDataBase, mm: ModelManager, idf_test: IdaDataFrame, 
               idf_train: IdaDataFrame, clean_up):
    model = ARule(idadb, MOD_NAME)
    assert model
    assert not mm.model_exists(MOD_NAME) 

    model.fit(idf_train, transaction_id_column='ID', item_column="PRODUCT", 
              support_type='absolute', support=2, confidence=0.4)
    assert mm.model_exists(MOD_NAME) 

    pred_ida = model.predict(idf_test, OUT_TABLE_PRED, max_conviction=1.25, 
                       max_affinity=0.6, min_lift=1.0, min_leverage=0.03)
    assert pred_ida
    assert all(pred_ida.columns == ['GID', 'TID', 'LHS_SID', 'RHS_SID', 'LHS_ITEMS', 'RHS_ITEMS', 'SUPPORT',
       'CONFIDENCE', 'LIFT', 'CONVICTION', 'AFFINITY', 'LEVERAGE'])
    pred = pred_ida.as_dataframe()
    assert all(pred["SUPPORT"].values >= 0.2)
    assert all(pred["CONFIDENCE"].values >= 0.4)
    assert all(pred["CONVICTION"].values <= 1.25)
    assert all(pred["AFFINITY"].values <= 0.6)
    assert all(pred["LIFT"].values >= 1.0)
    assert all(pred["LEVERAGE"].values >=0.03)

    assert model.describe()
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2023, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
import numpy as np
from nzpyida.base import IdaDataBase
from nzpyida.analytics.model_manager import ModelManager
from nzpyida.analytics.predictive.decision_trees import DecisionTreeClassifier
from nzpyida.analytics.tests.conftest import MOD_NAME, OUT_TABLE_PRED, OUT_TABLE_CM, TAB_NAME_TEST, \
    TAB_NAME_TRAIN, df_test, df_train
from nzpyida.analytics.tests.conftest import df_train, TAB_NAME_TRAIN, \
    TAB_NAME_TEST
from nzpyida.analytics import AutoDeleteContext

@pytest.fixture(scope='module')
def mm(idadb: IdaDataBase):
    return ModelManager(idadb)

@pytest.fixture(scope='module')
def clear_up(idadb: IdaDataBase, mm: ModelManager):
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CM):
        idadb.drop_table(OUT_TABLE_CM)
    yield
    if mm.model_exists(MOD_NAME):
        mm.drop_model(MOD_NAME)
    if idadb.exists_table(OUT_TABLE_PRED):
        idadb.drop_table(OUT_TABLE_PRED)
    if idadb.exists_table(OUT_TABLE_CM):
        idadb.drop_table(OUT_TABLE_CM)
    
def test_cross_validation(idadb: IdaDataBase, clear_up):
    idf = idadb.as_idadataframe(df_train, tablename=TAB_NAME_TRAIN, clear_existing=True, 
                                indexer="ID")

    model = DecisionTreeClassifier(idadb, MOD_NAME)
    with AutoDeleteContext(idadb):
        df_cv, acc = model.cross_validation(idf, id_column="ID", target_column="B")
        assert acc
        assert acc > 0.99

        assert df_cv
        assert len(df_cv) == len(idf)
        assert all(df_cv.columns == ["ID", "CLASS", "FOLD"])